#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2022 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~


import numpy as np
import pytest
import torch
from PIL import Image

from disent.util.visualize.vis_img import _ALLOWED_DTYPES
from disent.util.visualize.vis_img import _torch_to_images_normalise_args
from disent.util.visualize.vis_img import numpy_to_images
from disent.util.visualize.vis_img import numpy_to_pil_images
from disent.util.visualize.vis_img import torch_to_images

# ========================================================================= #
# Tests                                                                     #
# ========================================================================= #


def test_torch_to_images_basic():
    inp_float = torch.rand(8, 3, 64, 64, dtype=torch.float32)
    inp_uint8 = (inp_float * 127 + 63).to(torch.uint8)
    # check runs
    out = torch_to_images(inp_float)
    assert out.dtype == torch.uint8
    out = torch_to_images(inp_uint8)
    assert out.dtype == torch.uint8
    out = torch_to_images(inp_float, in_dtype=None, out_dtype=None)
    assert out.dtype == inp_float.dtype
    out = torch_to_images(inp_uint8, in_dtype=None, out_dtype=None)
    assert out.dtype == inp_uint8.dtype


def test_torch_to_images_permutations():
    inp_float = torch.rand(8, 3, 64, 64, dtype=torch.float32)
    inp_uint8 = (inp_float * 127 + 63).to(torch.uint8)

    # general checks
    def check_all(inputs, in_dtype=None):
        float_results, int_results = [], []
        for out_dtype in _ALLOWED_DTYPES:
            out = torch_to_images(inputs, in_dtype=in_dtype, out_dtype=out_dtype)
            stats = torch.stack(
                [out.min().to(torch.float64), out.max().to(torch.float64), out.to(dtype=torch.float64).mean()]
            )
            (float_results if out_dtype.is_floating_point else int_results).append(stats)
        for a, b in zip(float_results[:-1], float_results[1:]):
            assert torch.allclose(a, b)
        for a, b in zip(int_results[:-1], int_results[1:]):
            assert torch.allclose(a, b)

    # check type permutations
    check_all(inp_float, torch.float32)
    check_all(inp_uint8, torch.uint8)


def test_torch_to_images_preserve_type():
    for dtype in _ALLOWED_DTYPES:
        tensor = (torch.rand(8, 3, 64, 64) * (1 if dtype.is_floating_point else 255)).to(dtype)
        out = torch_to_images(tensor, in_dtype=dtype, out_dtype=dtype, clamp_mode="warn")
        assert out.dtype == dtype


def test_torch_to_images_arg_helper():
    assert _torch_to_images_normalise_args((64, 128, 3), torch.uint8, "HWC", "CHW", None, None) == (
        (-1, -3, -2),
        torch.uint8,
        torch.uint8,
        -3,
    )
    assert _torch_to_images_normalise_args((64, 128, 3), torch.uint8, "HWC", "HWC", None, None) == (
        (-3, -2, -1),
        torch.uint8,
        torch.uint8,
        -1,
    )


def test_torch_to_images_invalid_args():
    inp_float = torch.rand(8, 3, 64, 64, dtype=torch.float32)

    # check tensor
    with pytest.raises(TypeError, match="images must be of type"):
        torch_to_images(tensor=None)
    with pytest.raises(ValueError, match='dim "C", required: 1 or 3'):
        torch_to_images(tensor=torch.rand(8, 2, 16, 16, dtype=torch.float32))
    with pytest.raises(ValueError, match='dim "C", required: 1 or 3'):
        torch_to_images(tensor=torch.rand(8, 16, 16, 3, dtype=torch.float32))
    with pytest.raises(ValueError, match="images must have 3 or more dimensions corresponding to"):
        torch_to_images(tensor=torch.rand(16, 16, dtype=torch.float32))

    # check dims
    with pytest.raises(TypeError, match="in_dims must be of type"):
        torch_to_images(inp_float, in_dims=None)
    with pytest.raises(TypeError, match="out_dims must be of type"):
        torch_to_images(inp_float, out_dims=None)
    with pytest.raises(
        KeyError, match="in_dims contains the symbols: 'INVALID', must contain only permutations of: 'CHW'"
    ):
        torch_to_images(inp_float, in_dims="INVALID")
    with pytest.raises(
        KeyError, match="out_dims contains the symbols: 'INVALID', must contain only permutations of: 'CHW'"
    ):
        torch_to_images(inp_float, out_dims="INVALID")
    with pytest.raises(
        KeyError, match="in_dims contains the symbols: 'CHWW', must contain only permutations of: 'CHW'"
    ):
        torch_to_images(inp_float, in_dims="CHWW")
    with pytest.raises(
        KeyError, match="out_dims contains the symbols: 'CHWW', must contain only permutations of: 'CHW'"
    ):
        torch_to_images(inp_float, out_dims="CHWW")

    # check dtypes
    with pytest.raises(TypeError, match="images dtype: torch.float32 does not match in_dtype: torch.uint8"):
        torch_to_images(inp_float, in_dtype=torch.uint8)
    with pytest.raises(TypeError, match="in_dtype is not allowed"):
        torch_to_images(inp_float, in_dtype=torch.complex64)
    with pytest.raises(TypeError, match="out_dtype is not allowed"):
        torch_to_images(inp_float, out_dtype=torch.complex64)
    with pytest.raises(TypeError, match="in_dtype is not allowed"):
        torch_to_images(inp_float, in_dtype=torch.float16)
    with pytest.raises(TypeError, match="out_dtype is not allowed"):
        torch_to_images(inp_float, out_dtype=torch.float16)


def _check(target_shape, target_dtype, img, m=None, M=None):
    assert img.dtype == target_dtype
    assert img.shape == target_shape
    if isinstance(img, torch.Tensor):
        img = img.numpy()
    if m is not None:
        assert np.isclose(img.min(), m), f"min mismatch: {img.min()} (actual) != {m} (expected)"
    if M is not None:
        assert np.isclose(img.max(), M), f"max mismatch: {img.max()} (actual) != {M} (expected)"


def test_torch_to_images_adv():
    # CHW
    nchw_float = torch.rand(8, 3, 64, 32, dtype=torch.float32)
    nchw_uint8 = torch.randint(0, 255, (8, 3, 64, 32), dtype=torch.uint8)
    # HWC
    nhwc_float = torch.rand(8, 64, 32, 3, dtype=torch.float32)
    nhwc_uint8 = torch.randint(0, 255, (8, 64, 32, 3), dtype=torch.uint8)

    _check((8, 64, 32, 3), torch.uint8, torch_to_images(nchw_float))  # make sure default for numpy is CHW
    _check((8, 64, 32, 3), torch.uint8, torch_to_images(nchw_uint8))  # make sure default for numpy is CHW

    _check((8, 64, 32, 3), torch.uint8, torch_to_images(nchw_float, "CHW"))
    _check((8, 64, 32, 3), torch.uint8, torch_to_images(nchw_uint8, "CHW"))
    _check((8, 64, 32, 3), torch.uint8, torch_to_images(nhwc_float, "HWC"))
    _check((8, 64, 32, 3), torch.uint8, torch_to_images(nhwc_uint8, "HWC"))

    _check((8, 64, 32, 3), torch.uint8, torch_to_images(nchw_float, "CHW", "HWC"))
    _check((8, 64, 32, 3), torch.uint8, torch_to_images(nchw_uint8, "CHW", "HWC"))
    _check((8, 64, 32, 3), torch.uint8, torch_to_images(nhwc_float, "HWC", "HWC"))
    _check((8, 64, 32, 3), torch.uint8, torch_to_images(nhwc_uint8, "HWC", "HWC"))

    _check((8, 3, 64, 32), torch.uint8, torch_to_images(nchw_float, "CHW", "CHW"))
    _check((8, 3, 64, 32), torch.uint8, torch_to_images(nchw_uint8, "CHW", "CHW"))
    _check((8, 3, 64, 32), torch.uint8, torch_to_images(nhwc_float, "HWC", "CHW"))
    _check((8, 3, 64, 32), torch.uint8, torch_to_images(nhwc_uint8, "HWC", "CHW"))

    # random permute
    _check((8, 3, 64, 32), torch.uint8, torch_to_images(nchw_uint8, "CHW", "CHW"))
    _check((8, 3, 64, 32), torch.uint8, torch_to_images(nhwc_uint8, "HWC", "CHW"))
    _check((8, 3, 32, 64), torch.uint8, torch_to_images(nchw_uint8, "CHW", "CWH"))
    _check((8, 3, 32, 64), torch.uint8, torch_to_images(nhwc_uint8, "HWC", "CWH"))

    _check((8, 64, 3, 32), torch.uint8, torch_to_images(nchw_uint8, "CHW", "HCW"))
    _check((8, 64, 3, 32), torch.uint8, torch_to_images(nhwc_uint8, "HWC", "HCW"))
    _check((8, 32, 3, 64), torch.uint8, torch_to_images(nchw_uint8, "CHW", "WCH"))
    _check((8, 32, 3, 64), torch.uint8, torch_to_images(nhwc_uint8, "HWC", "WCH"))

    _check((8, 64, 32, 3), torch.uint8, torch_to_images(nchw_uint8, "CHW", "HWC"))
    _check((8, 64, 32, 3), torch.uint8, torch_to_images(nhwc_uint8, "HWC", "HWC"))
    _check((8, 32, 64, 3), torch.uint8, torch_to_images(nchw_uint8, "CHW", "WHC"))
    _check((8, 32, 64, 3), torch.uint8, torch_to_images(nhwc_uint8, "HWC", "WHC"))

    _check((8, 64, 32, 3), torch.float32, torch_to_images(nchw_float, "CHW", "HWC", out_dtype=torch.float32))
    _check((8, 64, 32, 3), torch.float32, torch_to_images(nchw_uint8, "CHW", "HWC", out_dtype=torch.float32))
    _check((8, 64, 32, 3), torch.float32, torch_to_images(nhwc_float, "HWC", "HWC", out_dtype=torch.float32))
    _check((8, 64, 32, 3), torch.float32, torch_to_images(nhwc_uint8, "HWC", "HWC", out_dtype=torch.float32))

    _check((8, 64, 32, 3), torch.float64, torch_to_images(nchw_float, "CHW", "HWC", out_dtype=torch.float64))
    _check((8, 64, 32, 3), torch.float64, torch_to_images(nchw_uint8, "CHW", "HWC", out_dtype=torch.float64))
    _check((8, 64, 32, 3), torch.float64, torch_to_images(nhwc_float, "HWC", "HWC", out_dtype=torch.float64))
    _check((8, 64, 32, 3), torch.float64, torch_to_images(nhwc_uint8, "HWC", "HWC", out_dtype=torch.float64))

    # random, but chance of this failing is almost impossible
    _check(
        (8, 64, 32, 3),
        torch.uint8,
        torch_to_images(nchw_float, "CHW", "HWC", out_dtype=torch.uint8, in_min=0.25, in_max=0.75),
        m=0,
        M=255,
    )
    _check(
        (8, 64, 32, 3),
        torch.uint8,
        torch_to_images(nchw_uint8, "CHW", "HWC", out_dtype=torch.uint8, in_min=64, in_max=192),
        m=0,
        M=255,
    )
    _check(
        (8, 64, 32, 3),
        torch.uint8,
        torch_to_images(nhwc_float, "HWC", "HWC", out_dtype=torch.uint8, in_min=0.25, in_max=0.75),
        m=0,
        M=255,
    )
    _check(
        (8, 64, 32, 3),
        torch.uint8,
        torch_to_images(nhwc_uint8, "HWC", "HWC", out_dtype=torch.uint8, in_min=64, in_max=192),
        m=0,
        M=255,
    )

    # random, but chance of this failing is almost impossible
    _check(
        (8, 64, 32, 3),
        torch.float32,
        torch_to_images(nchw_float, "CHW", "HWC", out_dtype=torch.float32, in_min=0.25, in_max=0.75),
        m=0,
        M=1,
    )
    _check(
        (8, 64, 32, 3),
        torch.float32,
        torch_to_images(nchw_uint8, "CHW", "HWC", out_dtype=torch.float32, in_min=64, in_max=192),
        m=0,
        M=1,
    )
    _check(
        (8, 64, 32, 3),
        torch.float32,
        torch_to_images(nhwc_float, "HWC", "HWC", out_dtype=torch.float32, in_min=0.25, in_max=0.75),
        m=0,
        M=1,
    )
    _check(
        (8, 64, 32, 3),
        torch.float32,
        torch_to_images(nhwc_uint8, "HWC", "HWC", out_dtype=torch.float32, in_min=64, in_max=192),
        m=0,
        M=1,
    )

    # random, but chance of this failing is almost impossible
    _check(
        (8, 3, 64, 32),
        torch.uint8,
        torch_to_images(nchw_float, "CHW", "CHW", out_dtype=torch.uint8, in_min=0.25, in_max=0.75),
        m=0,
        M=255,
    )
    _check(
        (8, 3, 64, 32),
        torch.uint8,
        torch_to_images(nchw_uint8, "CHW", "CHW", out_dtype=torch.uint8, in_min=64, in_max=192),
        m=0,
        M=255,
    )
    _check(
        (8, 3, 64, 32),
        torch.uint8,
        torch_to_images(nhwc_float, "HWC", "CHW", out_dtype=torch.uint8, in_min=0.25, in_max=0.75),
        m=0,
        M=255,
    )
    _check(
        (8, 3, 64, 32),
        torch.uint8,
        torch_to_images(nhwc_uint8, "HWC", "CHW", out_dtype=torch.uint8, in_min=64, in_max=192),
        m=0,
        M=255,
    )

    # random, but chance of this failing is almost impossible
    _check(
        (8, 3, 64, 32),
        torch.float32,
        torch_to_images(nchw_float, "CHW", "CHW", out_dtype=torch.float32, in_min=0.25, in_max=0.75),
        m=0,
        M=1,
    )
    _check(
        (8, 3, 64, 32),
        torch.float32,
        torch_to_images(nchw_uint8, "CHW", "CHW", out_dtype=torch.float32, in_min=64, in_max=192),
        m=0,
        M=1,
    )
    _check(
        (8, 3, 64, 32),
        torch.float32,
        torch_to_images(nhwc_float, "HWC", "CHW", out_dtype=torch.float32, in_min=0.25, in_max=0.75),
        m=0,
        M=1,
    )
    _check(
        (8, 3, 64, 32),
        torch.float32,
        torch_to_images(nhwc_uint8, "HWC", "CHW", out_dtype=torch.float32, in_min=64, in_max=192),
        m=0,
        M=1,
    )

    # check clamping
    with pytest.raises(ValueError, match="is outside of the required range"):
        torch_to_images(nchw_float, "CHW", out_dtype=torch.float32, clamp_mode="error", in_min=0.25, in_max=0.75)
    with pytest.raises(ValueError, match="is outside of the required range"):
        torch_to_images(nchw_uint8, "CHW", out_dtype=torch.float32, clamp_mode="error", in_min=64, in_max=192)
    with pytest.raises(ValueError, match="is outside of the required range"):
        torch_to_images(nhwc_float, "HWC", out_dtype=torch.float32, clamp_mode="error", in_min=0.25, in_max=0.75)
    with pytest.raises(ValueError, match="is outside of the required range"):
        torch_to_images(nhwc_uint8, "HWC", out_dtype=torch.float32, clamp_mode="error", in_min=64, in_max=192)
    with pytest.raises(KeyError, match="invalid clamp mode: 'asdf'"):
        torch_to_images(nhwc_uint8, "HWC", out_dtype=torch.float32, clamp_mode="asdf", in_min=64, in_max=192)


def test_numpy_to_pil_image():
    # CHW
    nchw_float = np.random.rand(8, 3, 64, 32)
    nchw_uint8 = np.random.randint(0, 255, (8, 3, 64, 32), dtype="uint8")

    # HWC
    nhwc_float = np.random.rand(8, 64, 32, 3)
    nhwc_uint8 = np.random.randint(0, 255, (8, 64, 32, 3), dtype="uint8")

    with pytest.raises(ValueError, match='images do not have the correct number of channels for dim "C"'):
        numpy_to_images(nchw_float)  # make sure default for numpy is HWC
    with pytest.raises(ValueError, match='images do not have the correct number of channels for dim "C"'):
        numpy_to_images(nchw_uint8)  # make sure default for numpy is HWC

    _check((8, 64, 32, 3), "uint8", numpy_to_images(nhwc_float))  # make sure default for numpy is HWC
    _check((8, 64, 32, 3), "uint8", numpy_to_images(nhwc_uint8))  # make sure default for numpy is HWC
    _check((8, 64, 32, 3), "uint8", numpy_to_images(nchw_float, "CHW"))
    _check((8, 64, 32, 3), "uint8", numpy_to_images(nchw_uint8, "CHW"))

    for pil_images in [
        numpy_to_pil_images(nchw_float, "CHW"),
        numpy_to_pil_images(nhwc_float),
    ]:
        assert isinstance(pil_images, np.ndarray)
        assert pil_images.shape == (8,)
        for pil_image in pil_images:
            pil_image: Image.Image
            assert isinstance(pil_image, Image.Image)
            assert pil_image.width == 32
            assert pil_image.height == 64

    # single image should be returned as an array of shape ()
    pil_image: np.ndarray = numpy_to_pil_images(np.random.rand(64, 32, 3))
    assert pil_image.shape == ()
    assert isinstance(pil_image, np.ndarray)
    assert isinstance(pil_image.tolist(), Image.Image)

    # check arb size
    pil_images: np.ndarray = numpy_to_pil_images(np.random.rand(4, 5, 2, 16, 32, 3))
    assert pil_images.shape == (4, 5, 2)


def test_numpy_image_min_max():
    # CHW
    nchw_float = np.random.rand(8, 3, 64, 32)
    nchw_uint8 = np.random.randint(0, 255, (8, 3, 64, 32), dtype="uint8")

    # HWC
    nhwc_float = np.random.rand(8, 64, 32, 3)
    nhwc_uint8 = np.random.randint(0, 255, (8, 64, 32, 3), dtype="uint8")

    _check((8, 64, 32, 3), "uint8", numpy_to_images(nhwc_float, "HWC", in_min=0, in_max=1))
    _check((8, 64, 32, 3), "uint8", numpy_to_images(nhwc_uint8, "HWC", in_min=0, in_max=255))
    _check((8, 64, 32, 3), "uint8", numpy_to_images(nchw_float, "CHW", in_min=0, in_max=1))
    _check((8, 64, 32, 3), "uint8", numpy_to_images(nchw_uint8, "CHW", in_min=0, in_max=255))

    # OUT: HWC

    _check((8, 64, 32, 3), "uint8", numpy_to_images(nhwc_float, "HWC", in_min=(0, 0, 0), in_max=(1, 1, 1)))
    _check((8, 64, 32, 3), "uint8", numpy_to_images(nhwc_uint8, "HWC", in_min=(0, 0, 0), in_max=(255, 255, 255)))
    _check((8, 64, 32, 3), "uint8", numpy_to_images(nchw_float, "CHW", in_min=(0, 0, 0), in_max=(1, 1, 1)))
    _check((8, 64, 32, 3), "uint8", numpy_to_images(nchw_uint8, "CHW", in_min=(0, 0, 0), in_max=(255, 255, 255)))

    _check(
        (8, 64, 32, 3), "uint8", numpy_to_images(nhwc_float, "HWC", in_min=(0,), in_max=(1,))
    )  # should maybe disable this from working?
    _check(
        (8, 64, 32, 3), "uint8", numpy_to_images(nhwc_uint8, "HWC", in_min=(0,), in_max=(255,))
    )  # should maybe disable this from working?
    _check(
        (8, 64, 32, 3), "uint8", numpy_to_images(nchw_float, "CHW", in_min=(0,), in_max=(1,))
    )  # should maybe disable this from working?
    _check(
        (8, 64, 32, 3), "uint8", numpy_to_images(nchw_uint8, "CHW", in_min=(0,), in_max=(255,))
    )  # should maybe disable this from working?

    _check((8, 64, 32, 1), "uint8", numpy_to_images(nhwc_float[:, :, :, 0:1], "HWC", in_min=(0,), in_max=(1,)))
    _check((8, 64, 32, 1), "uint8", numpy_to_images(nhwc_uint8[:, :, :, 0:1], "HWC", in_min=(0,), in_max=(255,)))
    _check((8, 64, 32, 1), "uint8", numpy_to_images(nchw_float[:, 0:1, :, :], "CHW", in_min=(0,), in_max=(1,)))
    _check((8, 64, 32, 1), "uint8", numpy_to_images(nchw_uint8[:, 0:1, :, :], "CHW", in_min=(0,), in_max=(255,)))

    _check((8, 64, 32, 1), "uint8", numpy_to_images(nhwc_float[:, :, :, 0:1], "HWC", in_min=0, in_max=1))
    _check((8, 64, 32, 1), "uint8", numpy_to_images(nhwc_uint8[:, :, :, 0:1], "HWC", in_min=0, in_max=255))
    _check((8, 64, 32, 1), "uint8", numpy_to_images(nchw_float[:, 0:1, :, :], "CHW", in_min=0, in_max=1))
    _check((8, 64, 32, 1), "uint8", numpy_to_images(nchw_uint8[:, 0:1, :, :], "CHW", in_min=0, in_max=255))

    # OUT: CHW

    _check((8, 3, 64, 32), "uint8", numpy_to_images(nhwc_float, "HWC", "CHW", in_min=(0, 0, 0), in_max=(1, 1, 1)))
    _check((8, 3, 64, 32), "uint8", numpy_to_images(nhwc_uint8, "HWC", "CHW", in_min=(0, 0, 0), in_max=(255, 255, 255)))
    _check((8, 3, 64, 32), "uint8", numpy_to_images(nchw_float, "CHW", "CHW", in_min=(0, 0, 0), in_max=(1, 1, 1)))
    _check((8, 3, 64, 32), "uint8", numpy_to_images(nchw_uint8, "CHW", "CHW", in_min=(0, 0, 0), in_max=(255, 255, 255)))

    _check(
        (8, 3, 64, 32), "uint8", numpy_to_images(nhwc_float, "HWC", "CHW", in_min=(0,), in_max=(1,))
    )  # should maybe disable this from working?
    _check(
        (8, 3, 64, 32), "uint8", numpy_to_images(nhwc_uint8, "HWC", "CHW", in_min=(0,), in_max=(255,))
    )  # should maybe disable this from working?
    _check(
        (8, 3, 64, 32), "uint8", numpy_to_images(nchw_float, "CHW", "CHW", in_min=(0,), in_max=(1,))
    )  # should maybe disable this from working?
    _check(
        (8, 3, 64, 32), "uint8", numpy_to_images(nchw_uint8, "CHW", "CHW", in_min=(0,), in_max=(255,))
    )  # should maybe disable this from working?

    _check((8, 1, 64, 32), "uint8", numpy_to_images(nhwc_float[:, :, :, 0:1], "HWC", "CHW", in_min=(0,), in_max=(1,)))
    _check((8, 1, 64, 32), "uint8", numpy_to_images(nhwc_uint8[:, :, :, 0:1], "HWC", "CHW", in_min=(0,), in_max=(255,)))
    _check((8, 1, 64, 32), "uint8", numpy_to_images(nchw_float[:, 0:1, :, :], "CHW", "CHW", in_min=(0,), in_max=(1,)))
    _check((8, 1, 64, 32), "uint8", numpy_to_images(nchw_uint8[:, 0:1, :, :], "CHW", "CHW", in_min=(0,), in_max=(255,)))

    _check((8, 1, 64, 32), "uint8", numpy_to_images(nhwc_float[:, :, :, 0:1], "HWC", "CHW", in_min=0, in_max=1))
    _check((8, 1, 64, 32), "uint8", numpy_to_images(nhwc_uint8[:, :, :, 0:1], "HWC", "CHW", in_min=0, in_max=255))
    _check((8, 1, 64, 32), "uint8", numpy_to_images(nchw_float[:, 0:1, :, :], "CHW", "CHW", in_min=0, in_max=1))
    _check((8, 1, 64, 32), "uint8", numpy_to_images(nchw_uint8[:, 0:1, :, :], "CHW", "CHW", in_min=0, in_max=255))


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

"""
Utilities for converting and testing different chunk sizes of hdf5 files
"""

import contextlib
import logging
import os
import warnings
from pathlib import Path
from typing import Any
from typing import Callable
from typing import Dict
from typing import Literal
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import Union

import h5py
import numpy as np
import torch
from torch.utils.data import DataLoader
from tqdm import tqdm

from disent.util.deprecate import deprecated
from disent.util.inout.files import AtomicSaveFile
from disent.util.iters import iter_chunks
from disent.util.profiling import Timer
from disent.util.strings import colors as c
from disent.util.strings.fmt import bytes_to_human

log = logging.getLogger(__name__)


# TODO: this file needs to be cleaned up!!!
# TODO: this file needs to be cleaned up!!!
# TODO: this file needs to be cleaned up!!!
# TODO: this file needs to be cleaned up!!!
# TODO: this file needs to be cleaned up!!!
# TODO: this file needs to be cleaned up!!!
# TODO: this file needs to be cleaned up!!!
# TODO: this file needs to be cleaned up!!!
# TODO: this file needs to be cleaned up!!!
# TODO: this file needs to be cleaned up!!!
# TODO: this file needs to be cleaned up!!!
# TODO: this file needs to be cleaned up!!!


# ========================================================================= #
# hdf5 - arguments                                                          #
# ========================================================================= #


AnyDType = Union[torch.dtype, np.dtype, str]


def _normalize_dtype(dtype: AnyDType) -> np.dtype:
    if isinstance(dtype, torch.dtype):
        dtype: str = torch.finfo(torch.float32).dtype
    return np.dtype(dtype)


ChunksType = Union[Tuple[int, ...], Literal["auto"], Literal["batch"]]


def _normalize_chunks(chunks: ChunksType, shape: Tuple[int, ...]):
    if chunks == "auto":
        warnings.warn("`auto` chuck shape has been used, did you mean `batch`")
        return True
    elif chunks == "batch":
        return (1, *shape[1:])
    elif isinstance(chunks, tuple):
        return chunks
    else:
        raise ValueError(f"invalid chunks value: {repr(chunks)}")


def _normalize_compression(compression_lvl: Optional[int]):
    if compression_lvl is None:
        return None, None  # compression, compression_lvl
    # check compression level
    if compression_lvl not in (0, 1, 2, 3, 4, 5, 6, 7, 8, 9):
        raise ValueError("compression_lvl must be an interger in the range [0, 9]")
    # get values
    return "gzip", compression_lvl  # compression, compression_lvl


# ========================================================================= #
# hdf5 - helper                                                             #
# ========================================================================= #


class H5IncompatibleError(Exception):
    pass


def h5_assert_deterministic(h5_file: h5py.File) -> h5py.File:
    # check the version
    if (isinstance(h5_file.libver, str) and h5_file.libver != "earliest") or (h5_file.libver[0] != "earliest"):
        raise H5IncompatibleError(
            f'hdf5 out file has an incompatible libver: {repr(h5_file.libver)} libver should be set to: "earliest"'
        )
    return h5_file


# ========================================================================= #
# hdf5 - resave                                                             #
# ========================================================================= #


@contextlib.contextmanager
def h5_open(path: str, mode: str = "r") -> h5py.File:
    """
    MODES:
        | atomic_w | Create temp file, then move and overwrite existing when done
        | atomic_x | Create temp file, then try move or fail if existing when done
        | r        | Readonly, file must exist (default)
        | r+       | Read/write, file must exist
        | w        | Create file, truncate if exists
        | w- or x  | Create file, fail if exists
        | a        | Read/write if exists, create otherwise
    """
    if not (str.endswith(path, ".h5") or str.endswith(path, ".hdf5")):
        warnings.warn(f"hdf5 file path does not end with extension: `.h5` or `.hdf5`, got: {path}")
    # get atomic context manager
    if mode == "atomic_w":
        save_context, mode = AtomicSaveFile(path, open_mode=None, overwrite=True), "w"
    elif mode == "atomic_x":
        save_context, mode = AtomicSaveFile(path, open_mode=None, overwrite=False), "x"
    else:
        save_context = contextlib.nullcontext(path)
    # handle saving to file
    with save_context as tmp_h5_path:
        with h5py.File(tmp_h5_path, mode, libver="earliest") as h5_file:
            yield h5_file


class H5Builder(object):
    def __init__(self, path: Union[str, Path], mode: str = "x"):
        """
        Supports the modes from `h5_open(...)`:
            | atomic_w | Create temp file, then move and overwrite existing when done
            | atomic_x | Create temp file, then try move or fail if existing when done
            | r        | Readonly, file must exist (default)
            | r+       | Read/write, file must exist
            | w        | Create file, truncate if exists
            | w- or x  | Create file, fail if exists
            | a        | Read/write if exists, create otherwise
        """
        super().__init__()
        # make sure that the file is deterministic
        # - we might be missing some of the properties that control this
        # - should we add a recursive option?
        if not isinstance(path, (str, Path)):
            raise TypeError(f"the given h5py path must be of type: `str`, `pathlib.Path`, got: {type(path)}")
        self._h5_path = path
        self._h5_mode = mode
        self._context_manager = None
        self._open_file = None

    def __enter__(self):
        self._context_manager = h5_open(self._h5_path, self._h5_mode)
        self._open_file = h5_assert_deterministic(self._context_manager.__enter__())
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._context_manager.__exit__(exc_type, exc_val, exc_tb)
        self._open_file = None
        self._context_manager = None

    @property
    def _h5_file(self) -> h5py.File:
        if self._open_file is None:
            raise "The H5Builder has not been opened in a new context, use `with H5Builder(...) as builder: ...`"
        return self._open_file

    def add_empty_dataset(
        self,
        name: str,
        shape: Tuple[int, ...],
        dtype: AnyDType,
        chunk_shape: ChunksType = "batch",
        compression_lvl: Optional[int] = 9,
        attrs: Optional[Dict[str, Any]] = None,
    ) -> "H5Builder":
        # normalize chunk_shape
        compression, compression_lvl = _normalize_compression(compression_lvl=compression_lvl)
        # create new dataset
        dataset = self._h5_file.create_dataset(
            name=name,
            shape=shape,
            dtype=_normalize_dtype(dtype),
            chunks=_normalize_chunks(chunk_shape, shape=shape),
            compression=compression,
            compression_opts=compression_lvl,
            # non-deterministic time stamps are added to the file if this is not
            # disabled, resulting in different hash sums when the file is re-generated!
            # - https://github.com/h5py/h5py/issues/225
            # - https://stackoverflow.com/questions/16019656
            # other properties:
            # - https://docs.h5py.org/en/stable/high/group.html#h5py.Group.create_dataset
            track_times=False,
            # how do these affect determinism:
            # track_order=False,
            # fletcher32=True,  # checksum for each chunk
            # shuffle=True,     # reorder chunk values to possibly help compression
            # scaleoffset=<int> # enable lossy compression, ints: number of bits to keep (0 is automatic lossless), floats: number of digits after decimal
        )
        # add attributes & convert
        if attrs is not None:
            for key, value in attrs.items():
                if isinstance(value, str):
                    value = np.array(value, dtype="S")
                dataset.attrs[key] = value
        # done!
        return self

    def fill_dataset(
        self,
        name: str,
        get_batch_fn: Callable[[int, int], np.ndarray],  # i_start, i_end
        batch_size: Union[int, Literal["auto"]] = "auto",
        show_progress: bool = False,
    ) -> "H5Builder":
        dataset: h5py.Dataset = self._h5_file[name]
        # determine batch size for copying data
        # get smallest multiple less than 32, otherwise original number
        if batch_size == "auto":
            if dataset.chunks:
                batch_size = dataset.chunks[0]
                batch_size = max((32 // batch_size) * batch_size, batch_size)
            else:
                batch_size = 32
        else:
            if dataset.chunks:
                if batch_size % dataset.chunks[0] != 0:
                    log.warning(
                        f"batch_size={batch_size} is not divisible by the first dimension of the dataset chunk size: {dataset.chunks[0]} {tuple(dataset.chunks)}"
                    )
        # check batch size!
        assert isinstance(batch_size, int) and (
            batch_size >= 1
        ), f'invalid batch_size: {repr(batch_size)}, expected: "auto" or an integer `>= 1`'
        # loop variables
        n = len(dataset)
        # save data
        with tqdm(total=n, disable=not show_progress, desc=f"saving {name}") as progress:
            for i in range(0, n, batch_size):
                j = min(i + batch_size, n)
                assert (
                    j > i
                ), f"this is a bug! {repr(j)} > {repr(i)}, len(dataset)={repr(n)}, batch_size={repr(batch_size)}"
                # load and modify the batch
                batch = get_batch_fn(i, j)  # i_start, i_end
                assert isinstance(batch, np.ndarray), f"returned batch is not an `np.ndarray`, got: {repr(type(batch))}"
                assert batch.shape == (
                    j - i,
                    *dataset.shape[1:],
                ), f"returned batch has incorrect shape: {tuple(batch.shape)}, expected: {(j-i, *dataset.shape[1:])}"
                # save the batch & update progress
                dataset[i:j] = batch
                progress.update(j - i)
        # done!
        return self

    def make_get_batch_fn(
        self,
        array,
        mutator: Optional[Callable[[np.ndarray], np.ndarray]] = None,
    ) -> Callable[[int, int], np.ndarray]:
        # get the array extractor
        if isinstance(array, torch.Tensor):

            def _extract_fn(i, j) -> np.ndarray:
                return array[i:j].detach().cpu().numpy()

        elif isinstance(array, (np.ndarray, h5py.Dataset)):  # chunk sizes will be missmatched

            def _extract_fn(i, j) -> np.ndarray:
                return array[i:j]

        elif isinstance(array, (tuple, list)):

            def _extract_fn(i, j) -> np.ndarray:
                return np.array(array[i:j])

        elif isinstance(array, Sequence):

            def _extract_fn(i, j) -> np.ndarray:
                return np.array([array[k] for k in range(i, j)])

        else:
            # last ditch effort, try as an iterator
            try:
                array = iter(array)
            except:
                raise TypeError(
                    f"`fill_dataset_from_array` only supports arrays of type: `np.ndarray` or `torch.Tensor`"
                )

            # get iterator function
            def _extract_fn(i, j) -> np.ndarray:
                return np.array([next(array) for k in range(i, j)])

        # make the get batch fn
        def get_batch_fn(i, j):
            batch: np.ndarray = _extract_fn(i, j)
            if mutator:
                batch = mutator(batch)
            assert isinstance(batch, np.ndarray), f"result of get_batch_fn is not a numpy array, got: {type(batch)}"
            return batch

        return get_batch_fn

    def fill_dataset_from_array(
        self,
        name: str,
        array,
        batch_size: Union[int, Literal["auto"]] = "auto",
        show_progress: bool = False,
        mutator: Optional[Callable[[np.ndarray], np.ndarray]] = None,
    ) -> "H5Builder":
        get_batch_fn = self.make_get_batch_fn(array, mutator=mutator)
        # get the batch size
        if batch_size == "auto" and isinstance(array, h5py.Dataset):
            batch_size = array.chunks[0]
        # copy into the dataset
        self.fill_dataset(
            name=name,
            get_batch_fn=get_batch_fn,
            batch_size=batch_size,
            show_progress=show_progress,
        )
        return self

    def fill_dataset_from_batches(
        self,
        name: str,
        batch_iter,
        batch_size: Union[int, Literal["auto"]] = "auto",
        show_progress: bool = False,
        mutator: Optional[Callable[[Any], np.ndarray]] = None,
    ) -> "H5Builder":
        try:
            batches = iter(batch_iter)
        except:
            raise TypeError(f"`fill_dataset_from_batches` must have iterable `batch_iter`, got: {type(batch_iter)}")

        # produce items
        def get_batch_fn(i, j):
            batch = next(batches)
            assert len(batch) == (j - i)
            if mutator:
                batch = mutator(batch)
            return np.array(batch)

        # copy into the dataset
        self.fill_dataset(
            name=name,
            get_batch_fn=get_batch_fn,
            batch_size=batch_size,
            show_progress=show_progress,
        )
        return self

    def add_dataset_from_array(
        self,
        array: np.ndarray,
        name: str = "data",
        chunk_shape: ChunksType = "batch",
        compression_lvl: Optional[int] = 4,
        attrs: Optional[Dict[str, Any]] = None,
        batch_size: Union[int, Literal["auto"]] = "auto",
        show_progress: bool = False,
        # optional, discovered automatically from array otherwise
        mutator: Optional[Callable[[np.ndarray], np.ndarray]] = None,
        dtype: Optional[np.dtype] = None,
        shape: Optional[Tuple[int, ...]] = None,
    ):
        self.add_empty_dataset(
            name=name,
            shape=array.shape if (shape is None) else shape,
            dtype=array.dtype if (dtype is None) else dtype,
            chunk_shape=chunk_shape,
            compression_lvl=compression_lvl,
            attrs=attrs,
        )
        self.fill_dataset_from_array(
            name=name,
            array=array,
            batch_size=batch_size,
            show_progress=show_progress,
            mutator=mutator,
        )

    def add_dataset_from_gt_data(
        self,
        data: Union["DisentDataset", "GroundTruthData"],
        name: str = "data",
        mutator: Optional[Callable[[Any], np.ndarray]] = None,
        img_shape: Tuple[Optional[int], ...] = (None, None, None),  # None items are automatically found
        batch_size: int = 32,
        compression_lvl: Optional[int] = 4,
        num_workers: int = min(os.cpu_count(), 16),
        show_progress: bool = True,
        chunk_shape: ChunksType = "batch",
        dtype: str = "uint8",
        attrs: Optional[dict] = None,
    ):
        from disent.dataset import DisentDataset
        from disent.dataset.data import GroundTruthData

        # get dataset
        # TODO: we should not automatically handle this extraction... The transform could be missing on the dataset?
        if isinstance(data, DisentDataset):
            gt_data = data.gt_data
        elif isinstance(data, GroundTruthData):
            gt_data = data
        else:
            raise TypeError(f"invalid data type: {type(data)}, must be {DisentDataset} or {GroundTruthData}")
        # process image shape
        H, W, C = img_shape
        if H is None:
            H = gt_data.img_shape[0]
        if W is None:
            W = gt_data.img_shape[1]
        if C is None:
            C = gt_data.img_shape[2]
        # make the empty dataset
        self.add_empty_dataset(
            name=name,
            shape=(len(gt_data), H, W, C),
            dtype=dtype,
            chunk_shape=chunk_shape,
            compression_lvl=compression_lvl,
            # THESE ATTRIBUTES SHOULD MATCH: SelfContainedHdf5GroundTruthData
            attrs=dict(
                dataset_name=gt_data.name,
                dataset_cls_name=gt_data.__class__.__name__,
                factor_sizes=np.array(gt_data.factor_sizes, dtype="uint"),
                factor_names=np.array(gt_data.factor_names, dtype="S"),
                # extra attrs -- we can't overwrite the above
                **(attrs if (attrs is not None) else {}),
            ),
        )
        # fill the dataset!
        self.fill_dataset_from_batches(
            name=name,
            batch_iter=DataLoader(
                gt_data, batch_size=batch_size, shuffle=False, num_workers=num_workers, drop_last=False
            ),
            batch_size=batch_size,
            show_progress=show_progress,
            mutator=mutator,
        )


#     def resave_dataset(self,
#         name: str,
#         inp: Union[str, Path, h5py.File, h5py.Dataset, np.ndarray],
#         # h5 re-save settings
#         chunk_shape: ChunksType = 'batch',
#         compression_lvl: Optional[int] = 4,
#         attrs: Optional[Dict[str, Any]] = None,
#         batch_size: Union[int, Literal['auto']] = 'auto',
#         show_progress: bool = False,
#         # optional, discovered automatically from array otherwise
#         mutator: Optional[Callable[[np.ndarray], np.ndarray]] = None,
#         dtype: Optional[np.dtype] = None,
#         obs_shape: Optional[Tuple[int, ...]] = None,
#     ):
#         # TODO: should this be more general and be able to handle add_dataset_from_gt_data too?
#         # TODO: this is very similar to save dataset below!
#         with _get_array_context(inp, name) as arr:
#             self.add_dataset_from_array(
#                 name=name,
#                 array=arr,
#                 chunk_shape=chunk_shape,
#                 compression_lvl=compression_lvl,
#                 attrs=attrs,
#                 batch_size=batch_size,
#                 show_progress=show_progress,
#                 mutator=mutator,
#                 dtype=dtype,
#                 shape=(len(arr), *obs_shape) if obs_shape else None,
#             )
#
#
# @contextlib.contextmanager
# def _get_array_context(
#     inp: Union[str, Path, h5py.File, h5py.Dataset, np.ndarray],
#     dataset_name: str = None,
# ) -> Union[h5py.Dataset, np.ndarray]:
#     # check the inputs
#     if not isinstance(inp, (str, Path, h5py.File, h5py.Dataset, np.ndarray)):
#         raise TypeError(f'unsupported input type: {type(inp)}')
#     # handle loading files
#     if isinstance(inp, str):
#         _, ext = os.path.splitext(inp)
#         if ext in ('.h5', '.hdf5'):
#             inp_context = h5py.File(inp, 'r')
#         else:
#             raise ValueError(f'unsupported extension: {repr(ext)} for path: {repr(inp)}')
#     else:
#         import contextlib
#         inp_context = contextlib.nullcontext(inp)
#     # re-save datasets
#     with inp_context as inp_data:
#         # get input dataset from h5 file
#         if isinstance(inp_data, h5py.File):
#             if dataset_name is None:
#                 raise ValueError('dataset_name must be specified if the input is an h5py.File so we can retrieve a h5py.Dataset')
#             inp_data = inp_data[dataset_name]
#         # return the data
#         yield inp_data


# ========================================================================= #
# hdf5 - resave                                                             #
# ========================================================================= #


# def _get_normalized_factor_names(gt_dataset, max_len=32):
#     # check inputs
#     if (factor_names is not None) and (factor_sizes is not None):
#         factor_names = tuple(s.encode('ascii') for s in factor_names)
#         factor_sizes = tuple(factor_sizes)
#         if any(len(s) > max_len for s in factor_names):
#             raise ValueError(f'factor names must be at most 32 ascii characters long')
#         if len(factor_names) != len(factor_sizes):
#             raise ValueError(f'length of factor names must be length of factor sizes: len({factor_names}) != len({factor_sizes})')
#         return factor_names, factor_sizes
#     elif not ((factor_names is None) and (factor_sizes is None)):
#         raise ValueError('factor_names and factor_sizes must both be given together.')
#     return None


def _normalize_out_array(array: Union[torch.Tensor, np.ndarray]) -> np.ndarray:
    if isinstance(array, torch.Tensor):
        return array.cpu().detach().numpy()
    return np.array(array)


@deprecated("this logic needs to be moved into `H5Builder`")
def hdf5_save_array(
    inp_data: Union[h5py.Dataset, np.ndarray, "torch.Tensor"],
    out_h5: h5py.File,
    dataset_name: str,  # input and output dataset name
    chunk_size: Optional[
        Union[Tuple[int, ...], Literal[True]]
    ] = None,  # True: auto determine, Tuple: specific chunk size, None: disable chunking
    compression: Optional[
        Union[Literal["gzip"], Literal["lzf"]]
    ] = None,  # compression type, only works if chunks is specified
    compression_lvl: Optional[int] = None,  # 0 through 9
    batch_size: Optional[int] = None,  # batch size to process / save at a time
    out_dtype: Optional[Union[np.dtype, str]] = None,  # output dtype of the dataset
    out_mutator: Optional[Callable[[np.ndarray], np.ndarray]] = None,  # mutate batches before saving
    obs_shape: Optional[Tuple[int, ...]] = None,  # resize batches to this shape
):
    # TODO: this should take in an array object and output the file!
    # check out_h5 version compatibility
    if (isinstance(out_h5.libver, str) and out_h5.libver != "earliest") or (out_h5.libver[0] != "earliest"):
        raise RuntimeError(
            f'hdf5 out file has an incompatible libver: {repr(out_h5.libver)} libver should be set to: "earliest"'
        )
    # get observation size
    if obs_shape is None:
        obs_shape = inp_data.shape[1:]
    # create new dataset
    out_data = out_h5.create_dataset(
        name=dataset_name,
        shape=(inp_data.shape[0], *obs_shape),
        dtype=out_dtype if (out_dtype is not None) else _normalize_dtype(inp_data.dtype),
        chunks=chunk_size,
        compression=compression,
        compression_opts=compression_lvl,
        # non-deterministic time stamps are added to the file if this is not
        # disabled, resulting in different hash sums when the file is re-generated!
        # - https://github.com/h5py/h5py/issues/225
        # - https://stackoverflow.com/questions/16019656
        # other properties:
        # - https://docs.h5py.org/en/stable/high/group.html#h5py.Group.create_dataset
        track_times=False,
        # track_order=False,
        # fletcher32=True,  # checksum for each chunk
        # shuffle=True,     # reorder chunk values to possibly help compression
        # scaleoffset=<int> # enable lossy compression, ints: number of bits to keep (0 is automatic lossless), floats: number of digits after decimal
    )
    # print stats
    tqdm.write("")
    hdf5_print_entry_data_stats(inp_data, label=f"IN")
    hdf5_print_entry_data_stats(out_data, label=f"OUT")
    # choose batch size for copying data
    if batch_size is None:
        batch_size = inp_data.chunks[0] if (hasattr(inp_data, "chunks") and inp_data.chunks) else 32
        log.debug(f"saving h5 dataset using automatic batch size of: {batch_size}")
    # get default
    if out_mutator is None:
        out_mutator = lambda x: x
    # save data
    with tqdm(total=len(inp_data)) as progress:
        for i in range(0, len(inp_data), batch_size):
            # load and modify the batch
            batch = inp_data[i : i + batch_size]
            batch = _normalize_out_array(batch)
            batch = out_mutator(batch)
            assert (
                batch.shape[1:] == obs_shape
            ), f"obs shape: {tuple(batch.shape[1:])} from processed input data does not match required obs shape: {tuple(obs_shape)}, try changing the `obs_shape` or resizing the batch in the `out_mutator`."
            # save the batch
            out_data[i : i + batch_size] = batch
            progress.update(batch_size)


@deprecated("this logic needs to be moved into `H5Builder`")
def hdf5_resave_file(
    inp_path: Union[str, torch.Tensor, np.ndarray],
    out_path: str,
    dataset_name: str,  # input and output dataset name
    chunk_size: Optional[
        Union[Tuple[int, ...], Literal[True]]
    ] = None,  # True: auto determine, Tuple: specific chunk size, None: disable chunking
    compression: Optional[
        Union[Literal["gzip"], Literal["lzf"]]
    ] = None,  # compression type, only works if chunks is specified
    compression_lvl: Optional[int] = None,  # 0 through 9
    batch_size: Optional[int] = None,  # batch size to process / save at a time
    out_dtype: Optional[Union[np.dtype, str]] = None,  # output dtype of the dataset
    out_mutator: Optional[Callable[[np.ndarray], np.ndarray]] = None,  # mutate batches before saving
    obs_shape: Optional[Tuple[int, ...]] = None,  # resize batches to this shape
    write_mode: Union[Literal["atomic_w"], Literal["w"], Literal["a"]] = "atomic_w",
):
    if isinstance(inp_path, str):
        inp_context = h5py.File(inp_path, "r")
    else:
        import contextlib

        inp_context = contextlib.nullcontext(inp_path)
    # re-save datasets
    with inp_context as inp_data:
        # get input dataset from h5 file
        if isinstance(inp_data, h5py.File):
            inp_data = inp_data[dataset_name]
        # get context manager
        if write_mode == "atomic_w":
            save_context = AtomicSaveFile(out_path, open_mode=None, overwrite=True)
            write_mode = "w"
        else:
            import contextlib

            save_context = contextlib.nullcontext(out_path)
        # handle saving to file
        with save_context as tmp_h5_path:
            with h5py.File(
                tmp_h5_path, write_mode, libver="earliest"
            ) as out_h5:  # TODO: libver='latest' is not deterministic, even with track_times=False
                hdf5_save_array(
                    inp_data=inp_data,
                    out_h5=out_h5,
                    dataset_name=dataset_name,
                    chunk_size=chunk_size,
                    compression=compression,
                    compression_lvl=compression_lvl,
                    batch_size=batch_size,
                    out_dtype=out_dtype,
                    out_mutator=out_mutator,
                    obs_shape=obs_shape,
                )
    # file size:
    log.info(
        f'[FILE SIZES] IN: {bytes_to_human(os.path.getsize(inp_path)) if isinstance(inp_path, str) else "N/A"} OUT: {bytes_to_human(os.path.getsize(out_path))}'
    )


# ========================================================================= #
# hdf5 - speed tests                                                        #
# ========================================================================= #


@deprecated("this logic needs to be moved into `H5Builder`")
def hdf5_test_entries_per_second(
    h5_dataset: h5py.Dataset, access_method="random", max_entries=48000, timeout=10, batch_size: int = 256
):
    # get access method
    if access_method == "sequential":
        indices = np.arange(len(h5_dataset))
    elif access_method == "random":
        indices = np.arange(len(h5_dataset))
        np.random.shuffle(indices)
    else:
        raise KeyError("Invalid access method")
    # num entries to test
    n = min(len(h5_dataset), max_entries)
    indices = indices[:n]
    # iterate through dataset, exit on timeout or max_entries
    t = Timer()
    for chunk in iter_chunks(enumerate(indices), chunk_size=batch_size):
        with t:
            for i, idx in chunk:
                entry = h5_dataset[idx]
        if t.elapsed > timeout:
            break
    # calculate score
    entries_per_sec = (i + 1) / t.elapsed
    return entries_per_sec


@deprecated("this logic needs to be moved into `H5Builder`")
def hdf5_test_speed(h5_path: str, dataset_name: str, access_method: str = "random"):
    with h5py.File(h5_path, "r") as out_h5:
        log.info("[TESTING] Access Speed...")
        log.info(
            f"Random Accesses Per Second: {hdf5_test_entries_per_second(out_h5[dataset_name], access_method=access_method, max_entries=5_000):.3f}"
        )


# ========================================================================= #
# hdf5 - stats                                                              #
# ========================================================================= #


@deprecated("this logic needs to be moved into `H5Builder`")
def hdf5_print_entry_data_stats(h5_dataset: h5py.Dataset, label="STATISTICS"):
    if not isinstance(h5_dataset, h5py.Dataset):
        tqdm.write(
            f"[{label:3s}] "
            f"array: {tuple(h5_dataset.shape)} ({str(h5_dataset.dtype):8s})"
            + (f" ({h5_dataset.device})" if isinstance(h5_dataset, torch.Tensor) else "")
        )
        return
    # get info
    itemsize = _normalize_dtype(h5_dataset.dtype).itemsize
    # entry
    shape = np.array([1, *h5_dataset.shape[1:]])
    data_per_entry = np.prod(shape) * itemsize
    # chunk
    chunks = np.array(h5_dataset.chunks) if (h5_dataset.chunks is not None) else np.ones(h5_dataset.ndim, dtype="int")
    data_per_chunk = np.prod(chunks) * itemsize
    # chunks per entry
    chunks_per_dim = np.ceil(shape / chunks).astype("int")
    chunks_per_entry = np.prod(chunks_per_dim)
    read_data_per_entry = data_per_chunk * chunks_per_entry
    # format
    chunks = f"{str(list(chunks)):18s}"
    data_per_chunk = f"{bytes_to_human(data_per_chunk):20s}"
    chunks_per_dim = f"{str(list(chunks_per_dim)):18s}"
    chunks_per_entry = f"{chunks_per_entry:5d}"
    read_data_per_entry = f"{bytes_to_human(read_data_per_entry):20s}"
    # format remaining
    entry = f"{str(list(shape)):18s}"
    data_per_entry = f"{bytes_to_human(data_per_entry)}"
    # print info
    tqdm.write(
        f"[{label:3s}] "
        f"entry: {entry} ({str(h5_dataset.dtype):8s}) {c.lYLW}{data_per_entry}{c.RST} "
        f"chunk: {chunks} {c.YLW}{data_per_chunk}{c.RST} "
        f"chunks per entry: {chunks_per_dim} {c.YLW}{read_data_per_entry}{c.RST} ({c.RED}{chunks_per_entry}{c.RST})  |  "
        f"compression: {repr(h5_dataset.compression)} compression lvl: {repr(h5_dataset.compression_opts)}"
    )


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

import logging

from disent.dataset.data._groundtruth import Hdf5GroundTruthData
from disent.dataset.util.datafile import DataFileHashedDlH5

# ========================================================================= #
# dataset_dsprites                                                          #
# ========================================================================= #


# TODO: this seems to have a memory leak compared to other datasets?
class DSpritesData(Hdf5GroundTruthData):
    """
    DSprites Dataset
    - beta-VAE: Learning Basic Visual Concepts with a Constrained Variational Framework
      (https://github.com/deepmind/dsprites-dataset)

    Files:
        - direct npz: https://raw.githubusercontent.com/deepmind/dsprites-dataset/master/dsprites_ndarray_co1sh3sc6or40x32y32_64x64.npz
                      approx 2.5 GB loaded into memory
        - direct hdf5: https://raw.githubusercontent.com/deepmind/dsprites-dataset/master/dsprites_ndarray_co1sh3sc6or40x32y32_64x64.hdf5
                       default chunk size is (23040, 2, 4), dataset is (737280, 64, 64) uint8.

    # reference implementation: https://github.com/google-research/disentanglement_lib/blob/master/disentanglement_lib/data/ground_truth/dsprites.py
    """

    name = "dsprites"

    # TODO: reference implementation has colour variants
    factor_names = ("shape", "scale", "orientation", "position_x", "position_y")
    factor_sizes = (3, 6, 40, 32, 32)  # TOTAL: 737280
    img_shape = (64, 64, 1)

    datafile = DataFileHashedDlH5(
        # download file/link
        uri="https://raw.githubusercontent.com/deepmind/dsprites-dataset/master/dsprites_ndarray_co1sh3sc6or40x32y32_64x64.hdf5",
        uri_hash={"fast": "d6ee1e43db715c2f0de3c41e38863347", "full": "b331c4447a651c44bf5e8ae09022e230"},
        # processed dataset file
        file_hash={"fast": "25013c85aebbf4b1023d72564f9413f0", "full": "4611d1a03e709cd5d0f6fdcdc221ca0e"},
        # h5 re-save settings
        hdf5_dataset_name="imgs",
        hdf5_chunk_size=(1, 64, 64, 1),
        hdf5_dtype="uint8",
        hdf5_mutator=lambda x: (x * 255)[..., None],  # data is of shape (-1, 64, 64), so we add the channel dimension
        hdf5_obs_shape=(64, 64, 1),
    )


# ========================================================================= #
# END                                                                       #
# ========================================================================= #


if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    DSpritesData(prepare=True)



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

from dataclasses import dataclass
from typing import Any
from typing import Dict
from typing import Sequence
from typing import Tuple

import torch
from torch.distributions import Distribution
from torch.distributions import Normal
from torch.distributions import kl_divergence

from disent.frameworks.vae._unsupervised__betavae import BetaVae

# ========================================================================= #
# Ada-GVAE                                                                  #
# ========================================================================= #


class AdaVae(BetaVae):
    """
    Weakly Supervised Disentanglement Learning Without Compromises: https://arxiv.org/abs/2002.02886
    - pretty much a beta-vae with averaging between decoder outputs to form weak supervision signal.
    - GAdaVAE:   Averaging from https://arxiv.org/abs/1809.02383
    - ML-AdaVAE: Averaging from https://arxiv.org/abs/1705.08841

    MODIFICATION:
    - Symmetric KL Calculation used by default, described in: https://arxiv.org/pdf/2010.14407.pdf
    - adjustable threshold value

    * This class is a little over complicated because it has the added functionality
      listed above, and because we want to re-use features elsewhere. The code can
      be compressed down into about ~20 neat lines for `hook_intercept_ds` if we
      select and chose fixed `cfg` values.
    """

    REQUIRED_OBS = 2

    @dataclass
    class cfg(BetaVae.cfg):
        ada_average_mode: str = "gvae"
        ada_thresh_mode: str = "symmetric_kl"  # kl, symmetric_kl, dist, sampled_dist
        ada_thresh_ratio: float = 0.5

    def hook_intercept_ds(
        self, ds_posterior: Sequence[Distribution], ds_prior: Sequence[Distribution]
    ) -> Tuple[Sequence[Distribution], Sequence[Distribution], Dict[str, Any]]:
        """
        Adaptive VAE Glue Method, putting the various components together
        1. find differences between deltas
        2. estimate a threshold for differences
        3. compute a shared mask from this threshold
        4. average together elements that should be considered shared

        (✓) Visual inspection against reference implementation:
            https://github.com/google-research/disentanglement_lib (aggregate_argmax)
        """
        d0_posterior, d1_posterior = ds_posterior
        # shared elements that need to be averaged, computed per pair in the batch.
        share_mask = self.compute_shared_mask_from_posteriors(
            d0_posterior, d1_posterior, thresh_mode=self.cfg.ada_thresh_mode, ratio=self.cfg.ada_thresh_ratio
        )
        # compute average posteriors
        new_ds_posterior = self.make_shared_posteriors(
            d0_posterior, d1_posterior, share_mask, average_mode=self.cfg.ada_average_mode
        )
        # return new args & generate logs
        return new_ds_posterior, ds_prior, {"shared": share_mask.sum(dim=1).float().mean()}

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # HELPER - POSTERIORS                                                   #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    @classmethod
    def compute_deltas_from_posteriors(cls, d0_posterior: Distribution, d1_posterior: Distribution, thresh_mode: str):
        """
        (✓) Visual inspection against reference implementation
        https://github.com/google-research/disentanglement_lib (compute_kl)
        - difference is that they don't multiply by 0.5 to get true kl, but that's not needed

        TODO: this might be numerically unstable with f32 passed to distributions
        """
        # shared elements that need to be averaged, computed per pair in the batch.
        # [𝛿_i ...]
        if thresh_mode == "kl":
            # ORIGINAL
            deltas = kl_divergence(d1_posterior, d0_posterior)
        elif thresh_mode == "symmetric_kl":
            # FROM: https://openreview.net/pdf?id=8VXvj1QNRl1
            kl_deltas_d1_d0 = kl_divergence(d1_posterior, d0_posterior)
            kl_deltas_d0_d1 = kl_divergence(d0_posterior, d1_posterior)
            deltas = (0.5 * kl_deltas_d1_d0) + (0.5 * kl_deltas_d0_d1)
        elif thresh_mode == "dist":
            deltas = cls.compute_deltas_from_zs(d1_posterior.mean, d0_posterior.mean)
        elif thresh_mode == "sampled_dist":
            deltas = cls.compute_deltas_from_zs(d1_posterior.rsample(), d0_posterior.rsample())
        else:
            raise KeyError(f"invalid thresh_mode: {repr(thresh_mode)}")

        # return values
        return deltas

    @classmethod
    def compute_shared_mask_from_posteriors(
        cls, d0_posterior: Distribution, d1_posterior: Distribution, thresh_mode: str, ratio=0.5
    ):
        return cls.estimate_shared_mask(
            z_deltas=cls.compute_deltas_from_posteriors(d0_posterior, d1_posterior, thresh_mode=thresh_mode),
            ratio=ratio,
        )

    @classmethod
    def make_shared_posteriors(
        cls, d0_posterior: Normal, d1_posterior: Normal, share_mask: torch.Tensor, average_mode: str
    ) -> Tuple[Normal, Normal]:
        # compute average posterior
        ave_posterior = AdaVae.compute_average_distribution(
            d0_posterior=d0_posterior, d1_posterior=d1_posterior, average_mode=average_mode
        )
        # select shared elements
        ave_d0_posterior = Normal(
            loc=torch.where(share_mask, ave_posterior.loc, d0_posterior.loc),
            scale=torch.where(share_mask, ave_posterior.scale, d0_posterior.scale),
        )
        ave_d1_posterior = Normal(
            loc=torch.where(share_mask, ave_posterior.loc, d1_posterior.loc),
            scale=torch.where(share_mask, ave_posterior.scale, d1_posterior.scale),
        )
        # return values
        return ave_d0_posterior, ave_d1_posterior

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # HELPER - MEAN/MU VALUES (same functionality as posterior versions)    #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    @classmethod
    def compute_deltas_from_zs(cls, z0: torch.Tensor, z1: torch.Tensor) -> torch.Tensor:
        return torch.abs(z0 - z1)

    @classmethod
    def compute_shared_mask_from_zs(cls, z0: torch.Tensor, z1: torch.Tensor, ratio: float = 0.5):
        return cls.estimate_shared_mask(z_deltas=cls.compute_deltas_from_zs(z0, z1), ratio=ratio)

    @classmethod
    def make_shared_zs(
        cls, z0: torch.Tensor, z1: torch.Tensor, share_mask: torch.Tensor
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        # compute average values
        ave = 0.5 * z0 + 0.5 * z1
        # select shared elements
        ave_z0 = torch.where(share_mask, ave, z0)
        ave_z1 = torch.where(share_mask, ave, z1)
        return ave_z0, ave_z1

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # HELPER - COMMON                                                       #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    @classmethod
    def estimate_shared_mask(cls, z_deltas: torch.Tensor, ratio: float) -> torch.Tensor:
        """
        Core of the adaptive VAE algorithm, estimating which factors
        have changed (or in this case which are shared and should remained unchanged
        by being be averaged) between pairs of observations.
        - custom ratio is an addition, when ratio==0.5 then
          this is equivalent to the original implementation.

        (✓) Visual inspection against reference implementation:
            https://github.com/google-research/disentanglement_lib (aggregate_argmax)
            - Implementation conversion is non-trivial, items are histogram binned.
              If we are in the second histogram bin, ie. 1, then kl_deltas <= kl_threshs
            - TODO: (aggregate_labels) An alternative mode exists where you can bind the
                    latent variables to any individual label, by one-hot encoding which
                    latent variable should not be shared: "enforce that each dimension
                    of the latent code learns one factor (dimension 1 learns factor 1)
                    and enforce that each factor of variation is encoded in a single
                    dimension."
        """
        assert 0 <= ratio <= 1, f"ratio must be in the range: 0 <= ratio <= 1, got: {repr(ratio)}"
        # threshold τ
        maximums = z_deltas.max(axis=1, keepdim=True).values  # (B, 1)
        minimums = z_deltas.min(axis=1, keepdim=True).values  # (B, 1)
        z_threshs = torch.lerp(minimums, maximums, weight=ratio)  # (B, 1)
        # true if 'unchanged' and should be average
        shared_mask = z_deltas < z_threshs  # broadcast (B, Z) and (B, 1) -> (B, Z)
        # return
        return shared_mask

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # HELPER - DISTRIBUTIONS                                                #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    @classmethod
    def compute_average_distribution(cls, d0_posterior: Normal, d1_posterior: Normal, average_mode: str) -> Normal:
        return _COMPUTE_AVE_FNS[average_mode](
            d0_posterior=d0_posterior,
            d1_posterior=d1_posterior,
        )


# ========================================================================= #
# Averaging Functions                                                       #
# ========================================================================= #


def compute_average_gvae_std(d0_posterior: Normal, d1_posterior: Normal) -> Normal:
    """
    Compute the arithmetic mean of the encoder distributions.
    - This is a custom function based on the Ada-GVAE averaging,
      except over the standard deviation instead of the variance!

    *NB* this is un-official!
    """
    assert isinstance(
        d0_posterior, Normal
    ), f"posterior distributions must be {Normal.__name__} distributions, got: {type(d0_posterior)}"
    assert isinstance(
        d1_posterior, Normal
    ), f"posterior distributions must be {Normal.__name__} distributions, got: {type(d1_posterior)}"
    # averages
    ave_std = 0.5 * (d0_posterior.stddev + d1_posterior.stddev)
    ave_mean = 0.5 * (d0_posterior.mean + d1_posterior.mean)
    # done!
    return Normal(loc=ave_mean, scale=ave_std)


def compute_average_gvae(d0_posterior: Normal, d1_posterior: Normal) -> Normal:
    """
    Compute the arithmetic mean of the encoder distributions.
    - Ada-GVAE Averaging function

    (✓) Visual inspection against reference implementation:
        https://github.com/google-research/disentanglement_lib (GroupVAEBase.model_fn)
    """
    assert isinstance(
        d0_posterior, Normal
    ), f"posterior distributions must be {Normal.__name__} distributions, got: {type(d0_posterior)}"
    assert isinstance(
        d1_posterior, Normal
    ), f"posterior distributions must be {Normal.__name__} distributions, got: {type(d1_posterior)}"
    # averages
    ave_var = 0.5 * (d0_posterior.variance + d1_posterior.variance)
    ave_mean = 0.5 * (d0_posterior.mean + d1_posterior.mean)
    # done!
    return Normal(loc=ave_mean, scale=torch.sqrt(ave_var))


def compute_average_ml_vae(d0_posterior: Normal, d1_posterior: Normal) -> Normal:
    """
    Compute the product of the encoder distributions.
    - Ada-ML-VAE Averaging function

    (✓) Visual inspection against reference implementation:
        https://github.com/google-research/disentanglement_lib (MLVae.model_fn)

    # TODO: recheck
    """
    assert isinstance(
        d0_posterior, Normal
    ), f"posterior distributions must be {Normal.__name__} distributions, got: {type(d0_posterior)}"
    assert isinstance(
        d1_posterior, Normal
    ), f"posterior distributions must be {Normal.__name__} distributions, got: {type(d1_posterior)}"
    # Diagonal matrix inverse: E^-1 = 1 / E
    # https://proofwiki.org/wiki/Inverse_of_Diagonal_Matrix
    z0_invvar, z1_invvar = d0_posterior.variance.reciprocal(), d1_posterior.variance.reciprocal()
    # average var: E^-1 = E1^-1 + E2^-1
    # disentanglement_lib: ave_var = 2 * z0_var * z1_var / (z0_var + z1_var)
    ave_var = 2 * (z0_invvar + z1_invvar).reciprocal()
    # average mean: u^T = (u1^T E1^-1 + u2^T E2^-1) E
    # disentanglement_lib: ave_mean = (z0_mean/z0_var + z1_mean/z1_var) * ave_var * 0.5
    ave_mean = (d0_posterior.mean * z0_invvar + d1_posterior.mean * z1_invvar) * ave_var * 0.5
    # done!
    return Normal(loc=ave_mean, scale=torch.sqrt(ave_var))


_COMPUTE_AVE_FNS = {
    "gvae": compute_average_gvae,
    "ml-vae": compute_average_ml_vae,
    "gvae_std": compute_average_gvae_std,  # this is un-official!
}


# ========================================================================= #
# Ada-GVAE                                                                  #
# ========================================================================= #


class AdaGVaeMinimal(BetaVae):
    """
    This is a direct implementation of the Ada-GVAE,
    which should be equivalent to the AdaVae with config values:

    >>> AdaVae.cfg(
    >>>    ada_average_mode='gvae',
    >>>    ada_thresh_mode='symmetric_kl',
    >>>    ada_thresh_ratio=0.5,
    >>> )
    """

    REQUIRED_OBS = 2

    def hook_intercept_ds(
        self, ds_posterior: Sequence[Distribution], ds_prior: Sequence[Distribution]
    ) -> Tuple[Sequence[Distribution], Sequence[Distribution], Dict[str, Any]]:
        """
        Adaptive VAE Method, putting the various components together
            1. compute differences between representations
            2. estimate a threshold for differences
            3. compute a shared mask from this threshold
            4. average together elements that are marked as shared

        (x) Visual inspection against reference implementation:
            https://github.com/google-research/disentanglement_lib (aggregate_argmax)
        """
        d0_posterior, d1_posterior = ds_posterior
        assert isinstance(
            d0_posterior, Normal
        ), f"posterior distributions must be {Normal.__name__} distributions, got: {type(d0_posterior)}"
        assert isinstance(
            d1_posterior, Normal
        ), f"posterior distributions must be {Normal.__name__} distributions, got: {type(d1_posterior)}"

        # [1] symmetric KL Divergence FROM: https://openreview.net/pdf?id=8VXvj1QNRl1
        z_deltas = 0.5 * kl_divergence(d1_posterior, d0_posterior) + 0.5 * kl_divergence(d0_posterior, d1_posterior)

        # [2] estimate threshold from deltas
        z_deltas_min = z_deltas.min(axis=1, keepdim=True).values  # (B, 1)
        z_deltas_max = z_deltas.max(axis=1, keepdim=True).values  # (B, 1)
        z_thresh = 0.5 * z_deltas_min + 0.5 * z_deltas_max  # (B, 1)

        # [3] shared elements that need to be averaged, computed per pair in the batch
        share_mask = z_deltas < z_thresh  # broadcast (B, Z) and (B, 1) to get (B, Z)

        # [4.a] compute average representations
        # - this is the only difference between the Ada-ML-VAE
        ave_mean = 0.5 * d0_posterior.mean + 0.5 * d1_posterior.mean
        ave_std = (0.5 * d0_posterior.variance + 0.5 * d1_posterior.variance) ** 0.5

        # [4.b] select shared or original values based on mask
        z0_mean = torch.where(share_mask, ave_mean, d0_posterior.loc)
        z1_mean = torch.where(share_mask, ave_mean, d1_posterior.loc)
        z0_std = torch.where(share_mask, ave_std, d0_posterior.scale)
        z1_std = torch.where(share_mask, ave_std, d1_posterior.scale)

        # construct distributions
        ave_d0_posterior = Normal(loc=z0_mean, scale=z0_std)
        ave_d1_posterior = Normal(loc=z1_mean, scale=z1_std)
        new_ds_posterior = (ave_d0_posterior, ave_d1_posterior)

        # [done] return new args & generate logs
        return new_ds_posterior, ds_prior, {"shared": share_mask.sum(dim=1).float().mean()}


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

import logging
import os
from pathlib import Path
from typing import Tuple
from typing import Union

log = logging.getLogger(__name__)


# ========================================================================= #
# Formatting                                                                #
# ========================================================================= #


def modify_file_name(file: Union[str, Path], prefix: str = None, suffix: str = None, sep=".") -> Union[str, Path]:
    # get path components
    path = Path(file)
    assert path.name, f"file name cannot be empty: {repr(path)}, for name: {repr(path.name)}"
    # create new path
    prefix = "" if (not prefix) else f"{prefix}{sep}"
    suffix = "" if (not suffix) else f"{sep}{suffix}"
    new_path = path.parent.joinpath(f"{prefix}{path.name}{suffix}")
    # return path with same format as input
    return str(new_path) if isinstance(file, str) else new_path


def modify_name_keep_ext(
    file: Union[str, Path], prefix: str = None, suffix: str = None, name_contains_sep: bool = False
) -> Union[str, Path]:
    # get path components
    path = Path(file)
    name = path.name
    assert name, f"file name cannot be empty: {repr(path)}, for name: {repr(name)}"
    # handle suffix
    if suffix:
        components = name.rsplit(".", 1) if name_contains_sep else path.name.split(".", 1)
        if len(components) >= 2:
            [name, ext] = components
            name = f"{name}{suffix}.{ext}"
        else:
            [name] = components
            name = f"{name}{suffix}"
    # handle prefix
    if prefix:
        name = f"{prefix}{name}"
    # create new path
    new_path = path.parent.joinpath(name)
    # return path with same format as input
    return str(new_path) if isinstance(file, str) else new_path


def modify_ext(file: Union[str, Path], ext: str, name_contains_sep: bool = True) -> Union[str, Path]:
    assert not ext.startswith("."), f"please specify the extension without the starting period: {repr(ext)}"
    # get path components
    path = Path(file)
    name = path.name
    assert name, f"file name cannot be empty: {repr(path)}, for name: {repr(name)}"
    # update the path name
    (name, *_) = name.rsplit(".", 1) if name_contains_sep else path.name.split(".", 1)
    name = f"{name}.{ext}"
    new_path = path.parent.joinpath(name)
    # return path with same format as input
    return str(new_path) if isinstance(file, str) else new_path


# ========================================================================= #
# files/dirs exist                                                          #
# ========================================================================= #


def ensure_dir_exists(*join_paths: str, is_file=False, absolute=False):
    import os

    # join path
    path = os.path.join(*join_paths)
    # to abs path
    if absolute:
        path = os.path.abspath(path)
    # remove file
    dirs = os.path.dirname(path) if is_file else path
    # create missing directory
    if os.path.exists(dirs):
        if not os.path.isdir(dirs):
            raise IOError(f"path is not a directory: {dirs}")
    else:
        os.makedirs(dirs, exist_ok=True)
        log.info(f"created missing directories: {dirs}")
    # return directory
    return path


def ensure_parent_dir_exists(*join_paths: str):
    return ensure_dir_exists(*join_paths, is_file=True, absolute=True)


# ========================================================================= #
# URI utils                                                                 #
# ========================================================================= #


def filename_from_url(url: str):
    import os
    from urllib.parse import urlparse

    return os.path.basename(urlparse(url).path)


def uri_parse_file_or_url(inp_uri: str) -> Tuple[str, bool]:
    from urllib.parse import urlparse

    result = urlparse(inp_uri)
    # parse different cases
    if result.scheme in ("http", "https"):
        is_url = True
        uri = result.geturl()
    elif result.scheme in ("file", ""):
        is_url = False
        if result.scheme == "file":
            if result.netloc:
                raise KeyError(
                    f'file uri format is invalid: "{result.geturl()}" two slashes specifies host as: "{result.netloc}" eg. instead of "file://hostname/root_folder/file.txt", please use: "file:/root_folder/file.txt" (no hostname) or "file:///root_folder/file.txt" (empty hostname).'
                )
            if not os.path.isabs(result.path):
                raise RuntimeError(
                    f"path: {repr(result.path)} obtained from file URI: {repr(inp_uri)} should always be absolute"
                )
            uri = result.path
        else:
            uri = result.geturl()
        uri = os.path.abspath(uri)
    else:
        raise ValueError(f"invalid file or url: {repr(inp_uri)}")
    # done
    return uri, is_url


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

import warnings
from typing import Optional
from typing import Tuple

import numpy as np

from disent.dataset.data._groundtruth import GroundTruthData

# ========================================================================= #
# helper                                                                    #
# ========================================================================= #


_R, _G, _B, _Y, _C, _M, _W = np.array(
    [
        [255, 000, 000],
        [000, 255, 000],
        [000, 000, 255],  # R, G, B
        [255, 255, 000],
        [000, 255, 255],
        [255, 000, 255],  # Y, C, M
        [255, 255, 255],  # white
    ]
)


def _shades(num: int, shades):
    all_shades = np.array([shade * i // num for i in range(1, num + 1) for shade in np.array(shades)])
    assert all_shades.dtype in ("int64", "int32")
    return all_shades


# ========================================================================= #
# xy object data                                                            #
# ========================================================================= #


class XYObjectData(GroundTruthData):
    """
    Michlo et al.
    https://github.com/nmichlo/msc-research

    Dataset that generates all possible permutations of a square placed on a square grid,
    with varying scale and colour

    *NB* for most of these color palettes, there should be
         an extra ground truth factor that represents shade.
         We purposely leave this out to hinder disentanglement! It is subjective!
    """

    name = "xy_object"

    COLOR_PALETTES_1 = {
        "greys_1": _shades(1, [[255]]),
        "greys_2": _shades(2, [[255]]),
        "greys_4": _shades(4, [[255]]),
        # aliases for greys so that we can just set `rgb=False` and it still works
        "rainbow_1": _shades(1, [[255]]),
        "rainbow_2": _shades(2, [[255]]),
        "rainbow_4": _shades(4, [[255]]),
    }

    COLOR_PALETTES_3 = {
        # grey
        "greys_1": _shades(1, [_W]),
        "greys_2": _shades(2, [_W]),
        "greys_4": _shades(4, [_W]),
        # colors -- white here and the incorrect ordering may throw off learning ground truth factors
        "colors_1": _shades(1, [_R, _G, _B, _Y, _C, _M, _W]),
        "colors_2": _shades(2, [_R, _G, _B, _Y, _C, _M, _W]),
        "colors_4": _shades(4, [_R, _G, _B, _Y, _C, _M, _W]),
        # rgb
        "rgb_1": _shades(1, [_R, _G, _B]),
        "rgb_2": _shades(2, [_R, _G, _B]),
        "rgb_4": _shades(4, [_R, _G, _B]),
        # rainbows -- these colors are mostly ordered correctly to align with gt factors
        "rainbow_1": _shades(1, [_R, _Y, _G, _C, _B, _M]),
        "rainbow_2": _shades(2, [_R, _Y, _G, _C, _B, _M]),
        "rainbow_4": _shades(4, [_R, _Y, _G, _C, _B, _M]),
    }

    factor_names = ("x", "y", "scale", "color")

    @property
    def factor_sizes(self) -> Tuple[int, ...]:
        return self._placements, self._placements, len(self._square_scales), len(self._colors)

    @property
    def img_shape(self) -> Tuple[int, ...]:
        return self._width, self._width, (3 if self._rgb else 1)

    def __init__(
        self,
        grid_size: int = 64,
        grid_spacing: int = 2,
        min_square_size: int = 7,
        max_square_size: int = 15,
        square_size_spacing: int = 2,
        rgb: bool = True,
        palette: str = "rainbow_4",
        transform=None,
    ):
        # generation
        self._rgb = rgb
        # check the pallete name
        assert (
            len(str.split(palette, "_")) == 2
        ), f"palette name must follow format: `<palette-name>_<brightness-levels>`, got: {repr(palette)}"
        # get the color palette
        color_palettes = XYObjectData.COLOR_PALETTES_3 if rgb else XYObjectData.COLOR_PALETTES_1
        if palette not in color_palettes:
            raise KeyError(
                f"color palette: {repr(palette)} does not exist for rgb={repr(rgb)}, select one of: {sorted(color_palettes.keys())}"
            )
        self._colors = color_palettes[palette]
        assert self._colors.ndim == 2
        assert self._colors.shape[-1] == (3 if rgb else 1)
        # image sizes
        self._width = grid_size
        # square scales
        assert min_square_size <= max_square_size
        self._max_square_size, self._max_square_size = min_square_size, max_square_size
        self._square_scales = np.arange(min_square_size, max_square_size + 1, square_size_spacing)
        # x, y
        self._spacing = grid_spacing
        self._placements = (self._width - max_square_size) // grid_spacing + 1
        super().__init__(transform=transform)

    def _get_observation(self, idx):
        x, y, s, c = self.idx_to_pos(idx)
        s = self._square_scales[s]
        r = (self._max_square_size - s) // 2
        x, y = self._spacing * x + r, self._spacing * y + r
        # GENERATE
        obs = np.zeros(self.img_shape, dtype=np.uint8)
        obs[y : y + s, x : x + s] = self._colors[c]
        return obs


# ========================================================================= #
# END                                                                       #
# ========================================================================= #


class XYObjectShadedData(XYObjectData):
    """
    Michlo et al.
    https://github.com/nmichlo/msc-research

    Dataset that generates all possible permutations of a square placed on a square grid,
    with varying scale and colour
    - This is like `XYObjectData` but has an extra factor that represents the shade.

    *NB* Changing the representation of the ground-truth factor can hurt or improve disentanglement performance.
         If distances in the data space along factors correlates better with ground-truth distances,
         then disentanglement performance is improved!
    """

    factor_names = ("x", "y", "scale", "intensity", "color")

    @property
    def factor_sizes(self) -> Tuple[int, ...]:
        return self._placements, self._placements, len(self._square_scales), self._brightness_levels, len(self._colors)

    @property
    def img_shape(self) -> Tuple[int, ...]:
        return self._width, self._width, (3 if self._rgb else 1)

    def __init__(
        self,
        grid_size: int = 64,
        grid_spacing: int = 2,
        min_square_size: int = 7,
        max_square_size: int = 15,
        square_size_spacing: int = 2,
        rgb: bool = True,
        palette: str = "rainbow_4",
        brightness_levels: Optional[int] = None,
        transform=None,
    ):
        parts = palette.split("_")
        if len(parts) > 1:
            # extract num levels from the string
            palette, b_levels = parts
            b_levels = int(b_levels)
            # handle conflict between brightness_levels and palette
            if brightness_levels is None:
                brightness_levels = b_levels
            else:
                warnings.warn(
                    f"palette ends with brightness_levels integer: {repr(b_levels)} (ignoring) but actual brightness_levels parameter was already specified: {repr(brightness_levels)} (using)"
                )
        # check the brightness_levels
        assert isinstance(
            brightness_levels, int
        ), f"brightness_levels must be an integer, got: {type(brightness_levels)}"
        assert 1 <= brightness_levels, f"brightness_levels must be >= 1, got: {repr(brightness_levels)}"
        self._brightness_levels = brightness_levels
        # initialize parent
        super().__init__(
            grid_size=grid_size,
            grid_spacing=grid_spacing,
            min_square_size=min_square_size,
            max_square_size=max_square_size,
            square_size_spacing=square_size_spacing,
            rgb=rgb,
            palette=f"{palette}_1",
            transform=transform,
        )

    def _get_observation(self, idx):
        x, y, s, b, c = self.idx_to_pos(idx)
        s = self._square_scales[s]
        r = (self._max_square_size - s) // 2
        x, y = self._spacing * x + r, self._spacing * y + r
        # GENERATE
        obs = np.zeros(self.img_shape, dtype=np.uint8)
        obs[y : y + s, x : x + s] = self._colors[c] * (b + 1) // self._brightness_levels
        return obs


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

from typing import Optional

import numpy as np

from disent.dataset.data import GroundTruthData
from disent.dataset.sampling._base import BaseDisentSampler
from disent.dataset.sampling._groundtruth__triplet import FactorSizeError
from disent.dataset.sampling._groundtruth__triplet import normalise_range_pair
from disent.dataset.util.state_space import StateSpace
from disent.util.math.random import sample_radius

# ========================================================================= #
# paired ground truth dataset                                               #
# ========================================================================= #


class GroundTruthPairSampler(BaseDisentSampler):
    def uninit_copy(self) -> "GroundTruthPairSampler":
        return GroundTruthPairSampler(
            p_k_range=self.p_k_range,
            p_radius_range=self.p_radius_range,
        )

    def __init__(
        self,
        # factor sampling
        p_k_range=(1, -1),
        # radius sampling
        p_radius_range=(1, -1),
    ):
        """
        Dataset that pairs together samples with at most k differing factors of variation.

        dataset: A dataset that extends GroundTruthData
        k: An integer (k), None (k=d-1), or "uniform" (random k in range 1 to d-1) -- inclusive of end value, not exclusive like np.random.randint
        variation_factor_indices: The indices of the factors of variation that are sampled between pairs, if None (all factors are sampled)
        """
        super().__init__(num_samples=2)
        self.p_k_range = p_k_range
        self.p_radius_range = p_radius_range
        # dataset variable
        self._state_space: Optional[StateSpace]

    def _init(self, dataset):
        assert isinstance(
            dataset, GroundTruthData
        ), f"dataset must be an instance of {repr(GroundTruthData.__class__.__name__)}, got: {repr(dataset)}"
        self._state_space = dataset.state_space_copy()
        # DIFFERING FACTORS
        self.p_k_min, self.p_k_max = self._min_max_from_range(
            p_range=self.p_k_range, max_values=self._state_space.num_factors
        )
        # RADIUS SAMPLING
        self.p_radius_min, self.p_radius_max = self._min_max_from_range(
            p_range=self.p_radius_range, max_values=self._state_space.factor_sizes
        )

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # CORE                                                                  #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    def _sample_idx(self, idx):
        f0, f1 = self.datapoint_sample_factors_pair(idx)
        return (
            self._state_space.pos_to_idx(f0),
            self._state_space.pos_to_idx(f1),
        )

    def datapoint_sample_factors_pair(self, idx):
        """
        Excerpt from Weakly-Supervised Disentanglement Without Compromises:
        [section 5. Experimental results]

        CREATE DATA SETS: with weak supervision from the existing
        disentanglement data sets:
        1. we first sample from the discrete z according to the ground-truth generative model (1)–(2).
        2. Then, we sample k factors of variation that should not be shared by the two images and re-sample those coordinates to obtain z˜.
           This ensures that each image pair differs in at most k factors of variation.

        For k we consider the range from 1 to d − 1.
        This last setting corresponds to the case where all but one factor of variation are re-sampled.

        We study both the case where k is constant across all pairs in the data set and where k is sampled uniformly in the range [d − 1] for every training pair (k = Rnd in the following).
        Unless specified otherwise, we aggregate the results for all values of k.
        """
        # SAMPLE FACTOR INDICES
        p_k = self._sample_num_factors()
        p_shared_indices = self._sample_shared_indices(p_k)
        # SAMPLE FACTORS - sample, resample and replace shared factors with originals
        anchor_factors = self._state_space.idx_to_pos(idx)
        positive_factors = self._resample_factors(anchor_factors)
        positive_factors[p_shared_indices] = anchor_factors[p_shared_indices]
        return anchor_factors, positive_factors

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # HELPER                                                                #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    def _min_max_from_range(self, p_range, max_values):
        p_min, p_max = normalise_range_pair(p_range, max_values)
        # cross factor assertions
        if not np.all(p_max <= max_values):
            raise FactorSizeError(
                "Factor dimensions are too small for given range:"
                f"\n\tUnsatisfied: p_max <= max_size"
                f"\n\tUnsatisfied: {p_max} <= {np.array(max_values)}"
            )
        return p_min, p_max

    def _sample_num_factors(self):
        p_k = np.random.randint(self.p_k_min, self.p_k_max + 1)
        return p_k

    def _sample_shared_indices(self, p_k):
        p_shared_indices = np.random.choice(
            self._state_space.num_factors, size=self._state_space.num_factors - p_k, replace=False
        )
        return p_shared_indices

    def _resample_factors(self, anchor_factors):
        positive_factors = sample_radius(
            anchor_factors,
            low=0,
            high=self._state_space.factor_sizes,
            r_low=self.p_radius_min,
            r_high=self.p_radius_max + 1,
        )
        return positive_factors


# ========================================================================= #
# END                                                                       #
# ========================================================================= #


# if __name__ == '__main__':
#     from disent.data.groundtruth import XYMultiGridData
#     from disent.util import concat_lines
#
#     # check that resample radius is working correctly!
#     dataset = GroundTruthDatasetPairs(
#         XYMultiGridData(1, 4),
#         p_k_range=(1, 1),
#         p_radius_range=(1, 1)
#     )
#
#     for pair in dataset:
#         obs0, obs1 = np.array(pair[0], dtype='int'), np.array(pair[1], dtype='int')
#         # CHECKS
#         diff = np.abs(obs1 - obs0)
#         diff_coords = np.array(np.where(diff > 0)).T
#         assert len(diff_coords) == 2  # check max changes
#         dist = np.abs(diff_coords[0] - diff_coords[1])
#         assert np.sum(dist > 0) == 1  # check max changes
#         assert np.max(dist) == 1      # check radius
#         # INFO
#         print(concat_lines(*[((obs > 0) * [1, 2, 4]).sum(axis=-1) for obs in (obs0, obs1)]), '\n')



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

import contextlib
import logging

log = logging.getLogger(__name__)


# ========================================================================= #
# seeds                                                                     #
# ========================================================================= #


def seed(long=777):
    """
    https://pytorch.org/docs/stable/notes/randomness.html
    """
    if long is None:
        log.warning(f"[SEEDING]: no seed was specified. Seeding skipped!")
        return
    # seed python
    import random

    random.seed(long)
    # seed numpy
    import numpy as np

    np.random.seed(long)
    # seed torch - it can be slow to import
    try:
        import torch

        torch.backends.cudnn.deterministic = True
        torch.backends.cudnn.benchmark = False
        torch.manual_seed(long)  # also calls: torch.cuda.manual_seed_all
    except ImportError:
        log.warning(f"[SEEDING]: torch is not installed. Skipped seeding torch methods!")
    # done!
    log.info(f"[SEEDED]: {long}")


class TempNumpySeed(contextlib.ContextDecorator):
    def __init__(self, seed: int = None):
        # check and normalize seed
        if seed is not None:
            try:
                seed = int(seed)
            except:
                raise ValueError(f"seed={seed} is not int-like!")
        # save values
        self._seed = seed
        self._state = None

    def __enter__(self):
        if self._seed is not None:
            import numpy as np

            self._state = np.random.get_state()
            np.random.seed(self._seed)

    def __exit__(self, *args, **kwargs):
        if self._seed is not None:
            import numpy as np

            np.random.set_state(self._state)
            self._state = None

    def _recreate_cm(self):
        # TODO: do we need to override this?
        return self


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~


import functools
from typing import Optional
from typing import Sequence

import numpy as np

# ========================================================================= #
# Dither Matrix                                                             #
# ========================================================================= #


def nd_dither_offsets(d: int) -> np.ndarray:
    """
    Get the offsets for the d-dimensional dither matrix.

    Output: array of shape: [2]*d

    Algorithm:
    M(d+1) = | 2 *     M(d)      |
             | 2 * flip(M(d)) + 1 |

    Examples:
    d=1:  | d=2:    | d=3:
    ------+---------+----------
    [0 1] | [[0 2]  | [[[0 4]
          |  [3 1]] |   [6 2]]
          |         |  [[3 7]
          |         |   [5 1]]]
    """
    assert isinstance(d, int) and (d > 0)
    # base offsets
    if d == 1:
        return np.array([0, 1])
    # recurse
    prev = nd_dither_offsets(d=d - 1)
    offs = np.array(
        [2 * prev, 2 * np.flip(prev) + 1]  # flip(prev) is the same as prev[::-1, >>>] with ::-1 in all dimensions
    )
    return _np_immutable_copy(offs)


def nd_dither_matrix(n: int = 2, d: int = 2, norm: bool = True) -> np.ndarray:
    """
    Compute the d-dimension dither matrix, with dimensions of size n.
    - n must be a power of 2!

    Output: array of shape: [n]*d

    Algorithm (d=2):
    M(2n) = | 4 * M(n) + 0     4 * M(n) + 2 |
            | 4 * M(n) + 3     4 * M(n) + 1 |

    Examples (n=2, norm=False):
    d=1:  | d=2:    | d=3:
    ------+---------+----------
    [0 1] | [[0 2]  | [[[0 4]
          |  [3 1]] |   [6 2]]
          |         |  [[3 7]
          |         |   [5 1]]]

    Examples (n=4, norm=False):
    d=1:      | d=2:
    ----------+----------------
    [0 2 1 3] | [[ 0  8  2 10]
              |  [12  4 14  6]
              |  [ 3 11  1  9]
              |  [15  7 13  5]]
    """
    assert _is_power_2(n)
    # handle smallest case
    if n == 1:
        return np.zeros([1] * d)  # shape: [1] * d
    # recurse
    offs = nd_dither_offsets(d=d)  # shape: [2] * d
    prev = nd_dither_matrix(n=n // 2, d=d, norm=False)  # shape: [N//2] * d
    # combine
    noffs = np.kron(
        offs, np.ones([n // 2] * d)
    )  # kron, eg. [0, 1] -> [0, 0, 1, 1] | we need to enlarge to shape: [n] * d
    nprev = np.tile(prev, [2] * d)  # tile, eg. [0, 1] -> [0, 1, 0, 1] | we need to enlarge to shape: [n] * d
    next = nprev * offs.size + noffs  # shape: [n] * d
    # return
    if norm:
        next /= next.size
    return _np_immutable_copy(next)  # shape: [n] * d


# ========================================================================= #
# Apply Dithering                                                           #
# ========================================================================= #


def nd_dither(arr: np.ndarray, n: int = 2, axis: Optional[Sequence[int]] = None) -> np.ndarray:
    """
    Apply ordered dithering along the specified axes of an array.
    The array must be floats with values in the range [0, 1]

    If axis is not specified, then all the axes are dithered.

    The output is a boolean array with the same shape as the input arr.
    """
    dd = nd_dither_matrix_like(arr, n=n, axis=axis, norm=True, expand=True)
    # compute dither
    return arr > dd


def nd_dither_matrix_like(
    arr: np.ndarray, n: int = 2, axis: Optional[Sequence[int]] = None, norm: bool = True, expand: bool = True
) -> np.ndarray:
    """
    Tile the dither matrix across an array.
    - `norm` specifies that the values should be in the range [0, 1)
       and not the original indices in the range [0, 2**d)
    - Use `axis` to specify which dimensions are tiled
      unspecified dimensions are set to size 1 for broadcasting
      with the original matrix, unless `expand=False`
    - `n` is the size of the underlying dither matrix which is tiled
    """
    axis = _normalize_axis(arr.ndim, tuple(axis))
    sizes = np.array(arr.shape)[axis]
    # get dither values
    d_mat = nd_dither_matrix(n=n, d=len(axis), norm=norm)
    # repeat values across array, rounding up and then trimming dims
    dd = np.tile(d_mat, (sizes + n - 1) // n)
    dd = dd[tuple(slice(0, l) for l in sizes)]
    # create missing dims
    if expand:
        dd = np.expand_dims(dd, axis=tuple(set(range(arr.ndim)) - set(axis)))
    # done
    return dd


# ========================================================================= #
# Helper                                                                    #
# ========================================================================= #


def _np_immutable_copy(arr: np.ndarray):
    arr = np.copy(arr)  # does not copy inner python objects
    arr.flags.writeable = False
    return arr


def _is_power_2(num: int):
    assert isinstance(num, int)
    if num <= 0:
        return False
    return not bool(num & (num - 1))


@functools.lru_cache()
def _normalize_axis(ndim: int, axis: Optional[Sequence[int]]) -> np.ndarray:
    # TODO: this functionality may be duplicated
    #       -- similar to np.normalize_axis_tuple(...)
    # defaults
    if axis is None:
        axis = np.arange(ndim)
    # convert
    axis = np.array(axis)
    if axis.ndim == 0:
        axis = axis[None]
    # checks
    assert axis.ndim == 1
    assert axis.dtype in ("int", "int32", "int64")
    # convert
    axis = np.where(axis < 0, ndim + axis, axis)
    axis = np.sort(axis)
    # checks
    assert np.unique(axis).shape == axis.shape
    assert np.all(0 <= axis)
    assert np.all(axis < ndim)
    # done!
    return _np_immutable_copy(axis)  # shape: [d]


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

from typing import Any
from typing import Dict
from typing import Sequence
from typing import Tuple

import torch

from disent.util.iters import aggregate_dict
from disent.util.iters import collect_dicts
from disent.util.iters import map_all

# ========================================================================= #
# AVE LOSS HELPER                                                           #
# ========================================================================= #


def detach_all(tensors: Sequence[torch.tensor], if_: bool = True):
    if if_:
        return tuple(tensor.detach() for tensor in tensors)
    return tensors


# ========================================================================= #
# AVE LOSS HELPER                                                           #
# ========================================================================= #


def compute_ave_loss(loss_fn, *arg_list, **common_kwargs) -> torch.Tensor:
    # compute all losses
    losses = map_all(loss_fn, *arg_list, collect_returned=False, common_kwargs=common_kwargs)
    # compute mean loss
    loss = torch.stack(losses).mean(dim=0)
    # return!
    return loss


def compute_ave_loss_and_logs(loss_and_logs_fn, *arg_list, **common_kwargs) -> Tuple[torch.Tensor, Dict[str, Any]]:
    # compute all losses
    losses, logs = map_all(loss_and_logs_fn, *arg_list, collect_returned=True, common_kwargs=common_kwargs)
    # compute mean loss
    loss = torch.stack(losses).mean(dim=0)
    # compute mean logs
    logs = aggregate_dict(collect_dicts(logs))
    # return!
    return loss, logs


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
import random
from typing import List
from typing import Literal
from typing import Union

import numpy as np
import pytest

from disent.dataset import DisentDataset
from disent.dataset.data import BaseEpisodesData
from disent.dataset.data import XYObjectData
from disent.dataset.sampling import *


class DummyEpisodesData(BaseEpisodesData):
    def _load_episode_observations(self) -> List[np.ndarray]:
        return [
            np.random.randn(57, 3, 64, 64),
            np.random.randn(42, 3, 64, 64),
            np.random.randn(26, 3, 64, 64),
            np.random.randn(97, 3, 64, 64),
        ]


@pytest.mark.parametrize(
    ["dataset", "num_samples", "check_mode", "sampler"],
    [
        [XYObjectData(), 1, "first", SingleSampler()],
        [XYObjectData(), 1, "first", GroundTruthSingleSampler()],
        # original AdaVAE sampling method
        [XYObjectData(), 2, "first", GroundTruthPairOrigSampler(p_k=1)],
        [XYObjectData(), 2, "first", GroundTruthPairOrigSampler(p_k=2)],
        [XYObjectData(), 2, "first", GroundTruthPairOrigSampler(p_k=-1)],
        # TODO: consider removing the pair sampler... it is difficult to maintain and confusing
        [XYObjectData(), 2, "first", GroundTruthPairSampler()],
        # TODO: consider removing the triplet sampler... it is difficult to maintain and confusing
        [XYObjectData(), 3, "first", GroundTruthTripleSampler()],
        [XYObjectData(), 3, "first", GroundTruthTripleSampler(swap_metric="k")],
        [XYObjectData(), 3, "first", GroundTruthTripleSampler(swap_metric="manhattan")],
        [XYObjectData(), 3, "first", GroundTruthTripleSampler(swap_metric="manhattan_norm")],
        [XYObjectData(), 3, "first", GroundTruthTripleSampler(swap_metric="euclidean")],
        [XYObjectData(), 3, "first", GroundTruthTripleSampler(swap_metric="euclidean_norm")],
        # [XYObjectData(), 3, 'first', GroundTruthTripleSampler(n_k_sample_mode='offset')],  # TODO: these are broken
        [XYObjectData(), 3, "first", GroundTruthTripleSampler(n_k_sample_mode="bounded_below")],
        [XYObjectData(), 3, "first", GroundTruthTripleSampler(n_k_sample_mode="random")],
        # [XYObjectData(), 3, 'first', GroundTruthTripleSampler(n_radius_sample_mode='offset')],  # TODO: these are broken
        [XYObjectData(), 3, "first", GroundTruthTripleSampler(n_radius_sample_mode="bounded_below")],
        [XYObjectData(), 3, "first", GroundTruthTripleSampler(n_radius_sample_mode="random")],
        [XYObjectData(), 1, "first", GroundTruthDistSampler(num_samples=1)],
        [XYObjectData(), 2, "first", GroundTruthDistSampler(num_samples=2)],
        [XYObjectData(), 3, "first", GroundTruthDistSampler(num_samples=3)],
        [XYObjectData(), 3, "first", GroundTruthDistSampler(num_samples=3, triplet_sample_mode="random")],
        [XYObjectData(), 3, "first", GroundTruthDistSampler(num_samples=3, triplet_sample_mode="factors")],
        [XYObjectData(), 3, "first", GroundTruthDistSampler(num_samples=3, triplet_sample_mode="manhattan")],
        [XYObjectData(), 3, "first", GroundTruthDistSampler(num_samples=3, triplet_sample_mode="manhattan_scaled")],
        [XYObjectData(), 3, "first", GroundTruthDistSampler(num_samples=3, triplet_sample_mode="combined")],
        [XYObjectData(), 3, "first", GroundTruthDistSampler(num_samples=3, triplet_sample_mode="combined_scaled")],
        [XYObjectData(), 1, "first", RandomSampler(num_samples=1)],
        [XYObjectData(), 2, "first", RandomSampler(num_samples=2)],
        [XYObjectData(), 3, "first", RandomSampler(num_samples=3)],
        [DummyEpisodesData(), 1, "any", RandomEpisodeSampler(num_samples=1)],
        [DummyEpisodesData(), 2, "any", RandomEpisodeSampler(num_samples=2)],
        [DummyEpisodesData(), 3, "any", RandomEpisodeSampler(num_samples=3)],
        [
            DummyEpisodesData(),
            3,
            "any",
            RandomEpisodeSampler(num_samples=3, sample_radius=3),
        ],  # sample_radius >= num_samples
        [DummyEpisodesData(), 3, "any", RandomEpisodeSampler(num_samples=3, sample_radius=4)],
        [DummyEpisodesData(), 3, "any", RandomEpisodeSampler(num_samples=3, sample_radius=-1)],
    ],
)
def test_samplers(
    dataset, num_samples: int, check_mode: Union[Literal["first"], Literal["any"]], sampler: BaseDisentSampler
):
    # check dataset
    wrapper = DisentDataset(dataset, sampler)
    assert len(wrapper) == len(dataset)
    assert sampler.num_samples == num_samples
    # check dataset init & samples
    for batch in wrapper:
        assert isinstance(batch, dict)
        assert len(batch["x_targ"]) == sampler.num_samples
        break

    # check sample
    def check_samples(i: int):
        indices = sampler(i)
        assert isinstance(indices, tuple)
        assert len(indices) == num_samples
        if check_mode == "first":
            assert i == indices[0]
        elif check_mode == "any":
            assert i in indices
        else:  # pragma: no cover
            raise RuntimeError("test mode is invalid!")

    # check indices
    check_samples(0)
    check_samples(len(dataset) - 1)
    for i in range(10):
        check_samples(random.randint(0, len(dataset) - 1))



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

import logging
from typing import Optional
from typing import Tuple
from typing import Union

import numpy as np

from disent.dataset.data._groundtruth import GroundTruthData
from disent.util.iters import iter_chunks

log = logging.getLogger(__name__)


# ========================================================================= #
# xy multi grid data                                                        #
# ========================================================================= #


class XYSquaresMinimalData(GroundTruthData):
    """
    Michlo et al.
    https://github.com/nmichlo/msc-research

    Dataset that generates all possible permutations of 3 (R, G, B) coloured
    squares placed on a square grid. This dataset is designed to not overlap
    in the reconstruction loss space.

    If you use this in your work, please cite: https://github.com/nmichlo/disent

    NOTE: Unlike XYSquaresData, XYSquaresMinimalData is the bare-minimum class
          to generate the same results as the default values for XYSquaresData,
          this class is a fair bit faster (~0.8x)!
          - All 3 squares are returned, in RGB, each square is size 8, with
            non-overlapping grid spacing set to 8 pixels, in total leaving
            8*8*8*8*8*8 factors. Images are uint8 with fill values 0 (bg)
            and 255 (fg).

    This dataset is adversarial in nature to auto-encoders that use pixel-wise reconstruction losses.
    - The AE or VAE cannot order observations in the latent space. Disentanglement performance suffers hugely.
    - This is because this dataset has constant distance between data-points as measured by the reconstruction loss.
      If the grid spacing is lowered, such that overlap is introduced, then disentanglement can occur again!
    """

    name = "xy_squares_minimal"

    @property
    def factor_names(self) -> Tuple[str, ...]:
        return "x_R", "y_R", "x_G", "y_G", "x_B", "y_B"

    @property
    def factor_sizes(self) -> Tuple[int, ...]:
        return 8, 8, 8, 8, 8, 8  # R, G, B squares

    @property
    def img_shape(self) -> Tuple[int, ...]:
        return 64, 64, 3

    def _get_observation(self, idx):
        # get factors
        factors = np.reshape(np.unravel_index(idx, self.factor_sizes), (-1, 2))
        # GENERATE
        obs = np.zeros(self.img_shape, dtype=np.uint8)
        for i, (fx, fy) in enumerate(factors):
            x, y = 8 * fx, 8 * fy
            obs[y : y + 8, x : x + 8, i] = 255
        return obs


# ========================================================================= #
# xy multi grid data                                                        #
# ========================================================================= #


class XYSquaresData(GroundTruthData):
    """
    Michlo et al.
    https://github.com/nmichlo/msc-research

    Dataset that generates all possible permutations of 3 (R, G, B) coloured
    squares placed on a square grid. This dataset is designed to not overlap
    in the reconstruction loss space. (if the spacing is set correctly.)

    If you use this in your work, please cite: https://github.com/nmichlo/disent

    NOTE: Unlike XYSquaresMinimalData, XYSquaresData allows adjusting various aspects
          of the data that is generated, but the generation process is slower (~1.25x).

    This dataset is adversarial in nature to auto-encoders that use pixel-wise reconstruction losses.
    - The AE or VAE cannot order observations in the latent space. Disentanglement performance suffers hugely.
    - This is because this dataset has constant distance between data-points as measured by the reconstruction loss.
      If the grid spacing is lowered, such that overlap is introduced, then disentanglement can occur again!
    """

    name = "xy_squares"

    @property
    def factor_names(self) -> Tuple[str, ...]:
        return ("x_R", "y_R", "x_G", "y_G", "x_B", "y_B")[: self._num_squares * 2]

    @property
    def factor_sizes(self) -> Tuple[int, ...]:
        return (self._placements, self._placements) * self._num_squares  # R, G, B squares

    @property
    def img_shape(self) -> Tuple[int, ...]:
        return self._width, self._width, (3 if self._rgb else 1)

    def __init__(
        self,
        square_size: int = 8,
        image_size: int = 64,
        grid_size: Optional[int] = None,
        grid_spacing: Optional[int] = None,
        num_squares: int = 3,
        rgb: bool = True,
        fill_value: Optional[Union[float, int]] = None,
        dtype: Union[np.dtype, str] = np.uint8,
        no_warnings: bool = False,
        transform=None,
    ):
        """
        :param square_size: the size of the individual squares in pixels
        :param image_size: the image size in pixels
        :param grid_spacing: the step size between square positions on the grid. By
               default this is set to square_size which results in non-overlapping
               data if `grid_spacing >= square_size` Reducing this value such that
               `grid_spacing < square_size` results in overlapping data.
        :param num_squares: The number of squares drawn. `1 <= num_squares <= 3`
        :param rgb: Image has 3 channels if True, otherwise it is greyscale with 1 channel.
        :param no_warnings: If warnings should be disabled if overlapping.
        :param fill_value: The foreground value to use for filling squares, the default background value is 0.
        :param grid_size: The number of grid positions available for the square to be placed in. The square is centered if this is less than
        :param dtype:
        """
        if grid_spacing is None:
            grid_spacing = square_size
        if (grid_spacing < square_size) and not no_warnings:
            log.warning(f"overlap between squares for reconstruction loss, {grid_spacing} < {square_size}")
        # color
        self._rgb = rgb
        self._dtype = np.dtype(dtype)
        # check fill values
        if self._dtype.kind == "u":
            self._fill_value = 255 if (fill_value is None) else fill_value
            assert isinstance(self._fill_value, int)
            assert 0 < self._fill_value <= 255, f"0 < {self._fill_value} <= 255"
        elif self._dtype.kind == "f":
            self._fill_value = 1.0 if (fill_value is None) else fill_value
            assert isinstance(self._fill_value, (int, float))
            assert 0.0 < self._fill_value <= 1.0, f"0.0 < {self._fill_value} <= 1.0"
        else:
            raise TypeError(f"invalid dtype: {self._dtype}, must be float or unsigned integer")
        # image sizes
        self._width = image_size
        # number of squares
        self._num_squares = num_squares
        assert 1 <= num_squares <= 3, "Only 1, 2 or 3 squares are supported!"
        # square scales
        self._square_size = square_size
        # x, y
        self._spacing = grid_spacing
        self._placements = (self._width - self._square_size) // grid_spacing + 1
        # maximum placements
        if grid_size is not None:
            assert isinstance(grid_size, int)
            assert grid_size > 0
            if (grid_size > self._placements) and not no_warnings:
                log.warning(
                    f"number of possible placements: {self._placements} is less than the given grid size: {grid_size}, reduced grid size from: {grid_size} -> {self._placements}"
                )
            self._placements = min(self._placements, grid_size)
        # center elements
        self._offset = (self._width - (self._square_size + (self._placements - 1) * self._spacing)) // 2
        # initialise parents -- they depend on self.factors
        super().__init__(transform=transform)

    def _get_observation(self, idx):
        # get factors
        factors = self.idx_to_pos(idx)
        offset, space, size = self._offset, self._spacing, self._square_size
        # GENERATE
        obs = np.zeros(self.img_shape, dtype=self._dtype)
        for i, (fx, fy) in enumerate(iter_chunks(factors, 2)):
            x, y = offset + space * fx, offset + space * fy
            if self._rgb:
                obs[y : y + size, x : x + size, i] = self._fill_value
            else:
                obs[y : y + size, x : x + size, :] = self._fill_value
        return obs


# ========================================================================= #
# xy minimal single square dataset                                          #
# ========================================================================= #


class XYSingleSquareData(GroundTruthData):
    """
    Michlo et al.
    https://github.com/nmichlo/msc-research

    This is a very basic version of XYSquares that only has a
    single object that moves around, instead of 3.
    - This dataset is still adversarial if the grid_spacing is not adjusted.
    """

    name = "xy_single_square"

    factor_names = ("x", "y")

    @property
    def factor_sizes(self) -> Tuple[int, ...]:
        return (self._placements, self._placements)

    @property
    def img_shape(self) -> Tuple[int, ...]:
        return self._width, self._width, 1

    def __init__(
        self,
        square_size: int = 8,  # square width and height
        image_size: int = 64,  # image width and height
        grid_size: Optional[
            int
        ] = None,  # limit the number of square placements along an axis, automatically set as the maximum valid
        grid_spacing: Optional[
            int
        ] = None,  # how far apart the square is spaced, buy default this is the square size, meaning no overlap!
        no_warnings: bool = False,
        transform=None,
    ):
        if grid_spacing is None:
            grid_spacing = square_size
        if (grid_spacing < square_size) and not no_warnings:
            log.warning(f"overlap between squares for reconstruction loss, {grid_spacing} < {square_size}")
        # vars
        self._width = image_size  # image width and height
        self._square_size = square_size  # square width and height
        self._spacing = grid_spacing  # spacing between square positions
        self._placements = (
            self._width - self._square_size
        ) // grid_spacing + 1  # number of positions the square can be in along an axis
        # maximum placements
        if grid_size is not None:
            if grid_size > self._placements:
                log.warning(
                    f"number of possible placements: {self._placements} is less than the given grid size: {grid_size}, reduced grid size from: {grid_size} -> {self._placements}"
                )
            self._placements = min(self._placements, grid_size)
        # center elements
        self._offset = (self._width - (self._square_size + (self._placements - 1) * self._spacing)) // 2
        # initialise parents -- they depend on self.factors
        super().__init__(transform=transform)

    def _get_observation(self, idx):
        # get factors == grid position/index
        fx, fy = self.idx_to_pos(idx)
        offset, space, size = self._offset, self._spacing, self._square_size
        # draw square onto image
        obs = np.zeros(self.img_shape, dtype="uint8")
        x, y = offset + space * fx, offset + space * fy
        obs[y : y + size, x : x + size, :] = 255
        return obs


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

import torch

# ========================================================================= #
# PCA                                                                       #
# ========================================================================= #


def torch_pca_eig(X, center=True, scale=False, zero_negatives=False):
    """
    perform PCA over X
    - X is of size (num_points, vec_size)

    NOTE: unlike PCA_svd, the number of vectors/values returned is always: vec_size

    WARNING: this may be incorrect!
    """
    n, m = X.shape
    # center points along axes
    if center:
        X = X - X.mean(dim=0)
    # compute covariance -- TODO: optimise this line
    covariance = (1 / (n - 1)) * torch.mm(X.T, X)
    if scale:
        scaling = torch.sqrt(1 / torch.diagonal(covariance))
        covariance = torch.mm(torch.diagflat(scaling), covariance)
    # compute eigen values and eigen vectors
    eigenvalues, eigenvectors = torch.linalg.eig(covariance)
    # sort components by decreasing variance
    components = torch.real(eigenvectors.T)  # TODO: handle imaginary numbers!
    explained_variance = torch.real(eigenvalues)  # TODO: handle imaginary numbers!
    # handle n < m -- numerical stability issues return negative values!
    #                 maybe this should just zero out the negatives instead, they don't contribute!?
    explained_variance = torch.abs(explained_variance)
    # return sorted
    idxs = torch.argsort(explained_variance, descending=True)
    return components[idxs], explained_variance[idxs]


def torch_pca_svd(X, center=True):
    """
    perform PCA over X
    - X is of size (num_points, vec_size)

    NOTE: unlike PCA_eig, the number of vectors/values returned is: min(num_points, vec_size)
    """
    n, _ = X.shape
    # center points along axes
    if center:
        X = X - X.mean(dim=0)
    # perform singular value decomposition
    u, s, v = torch.svd(X)
    # sort components by decreasing variance
    # these are already sorted?
    components = v.T
    explained_variance = torch.mul(s, s) / (n - 1)
    return components, explained_variance


def torch_pca(X, center=True, mode="svd") -> (torch.Tensor, torch.Tensor):
    # number of values returned may differ depending on the method!
    # -- svd returns: min(num, z_size)
    # -- eig returns: num
    if mode == "svd":
        return torch_pca_svd(X, center=center)
    elif mode == "eig":
        return torch_pca_eig(X, center=center, scale=False)
    else:
        raise KeyError(f"invalid torch_pca mode: {repr(mode)}")


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

import numpy as np
import torch

from disent.nn.functional._other import torch_unsqueeze_l
from disent.nn.functional._other import torch_unsqueeze_r
from disent.nn.functional._util_generic import TypeGenericTensor
from disent.nn.functional._util_generic import TypeGenericTorch
from disent.nn.functional._util_generic import generic_as_int32
from disent.nn.functional._util_generic import generic_max

# ========================================================================= #
# Kernels                                                                   #
# ========================================================================= #


# TODO: replace with meshgrid based functions from experiment/exp/06_metric
#       these are more intuitive and flexible


def get_kernel_size(sigma: TypeGenericTensor = 1.0, truncate: TypeGenericTensor = 4.0):
    """
    This is how sklearn chooses kernel sizes.
    - sigma is the standard deviation, and truncate is the number of deviations away to truncate
    - our version broadcasts sigma and truncate together, returning the max kernel size needed over all values
    """
    # compute radius
    radius = generic_as_int32(truncate * sigma + 0.5)
    # get maximum value
    radius = int(generic_max(radius))
    # compute diameter
    return 2 * radius + 1


def torch_gaussian_kernel(
    sigma: TypeGenericTorch = 1.0,
    truncate: TypeGenericTorch = 4.0,
    size: int = None,
    dtype=torch.float32,
    device=None,
):
    # broadcast tensors together -- data may reference single memory locations
    sigma = torch.as_tensor(sigma, dtype=dtype, device=device)
    truncate = torch.as_tensor(truncate, dtype=dtype, device=device)
    sigma, truncate = torch.broadcast_tensors(sigma, truncate)
    # compute default size
    if size is None:
        size: int = get_kernel_size(sigma=sigma, truncate=truncate)
    # compute kernel
    x = torch.arange(size, dtype=sigma.dtype, device=sigma.device) - (size - 1) / 2
    # pad tensors correctly
    x = torch_unsqueeze_l(x, n=sigma.ndim)
    s = torch_unsqueeze_r(sigma, n=1)
    # compute
    return torch.exp(-(x**2) / (2 * s**2)) / (np.sqrt(2 * np.pi) * s)


def torch_gaussian_kernel_2d(
    sigma: TypeGenericTorch = 1.0,
    truncate: TypeGenericTorch = 4.0,
    size: int = None,
    sigma_b: TypeGenericTorch = None,
    truncate_b: TypeGenericTorch = None,
    size_b: int = None,
    dtype=torch.float32,
    device=None,
):
    # set default values
    if sigma_b is None:
        sigma_b = sigma
    if truncate_b is None:
        truncate_b = truncate
    if size_b is None:
        size_b = size
    # compute kernel
    kh = torch_gaussian_kernel(sigma=sigma, truncate=truncate, size=size, dtype=dtype, device=device)
    kw = torch_gaussian_kernel(sigma=sigma_b, truncate=truncate_b, size=size_b, dtype=dtype, device=device)
    return kh[..., :, None] * kw[..., None, :]


def torch_box_kernel(radius: TypeGenericTorch = 1, dtype=torch.float32, device=None):
    radius = torch.abs(torch.as_tensor(radius, device=device))
    assert radius.dtype in {torch.int32, torch.int64}, f"box kernel radius must be of integer type: {radius.dtype}"
    # box kernel values
    radius_max = radius.max()
    crange = torch.abs(torch.arange(radius_max * 2 + 1, dtype=dtype, device=device) - radius_max)
    # pad everything
    radius = radius[..., None]
    crange = crange[None, ...]
    # compute box kernel
    kernel = (crange <= radius).to(dtype) / (radius * 2 + 1)
    # done!
    return kernel


def torch_box_kernel_2d(
    radius: TypeGenericTorch = 1, radius_b: TypeGenericTorch = None, dtype=torch.float32, device=None
):
    # set default values
    if radius_b is None:
        radius_b = radius
    # compute kernel
    kh = torch_box_kernel(radius=radius, dtype=dtype, device=device)
    kw = torch_box_kernel(radius=radius_b, dtype=dtype, device=device)
    return kh[..., :, None] * kw[..., None, :]


# ========================================================================= #
# END                                                                       #
# ========================================================================= #
